# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from knext.builder import rest
from knext.builder.component.base import SinkWriter
from knext.builder.operator.spg_record import SPGRecord


class KGWriter(SinkWriter):
    """The Sink Component that writing data to KG storage.

    Args:
        None
    Examples:
        sink = KGWriter()

    """

    @property
    def input_types(self):
        return SPGRecord

    @property
    def output_types(self):
        return None

    def to_rest(self):
        """Transforms `KGWriter` to REST model `GraphStoreSinkNodeConfig`."""
        config = rest.GraphStoreSinkNodeConfig()
        return rest.Node(**super().to_dict(), node_config=config)

    @classmethod
    def from_rest(cls, node: rest.Node):
        pass

    def invoke(self, input):
        pass

    def submit(self):
        pass



# coding: utf-8
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

"""
    knext

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from knext.common.rest.configuration import Configuration


class BaseLinkingConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {"strategy_type": "str", "linking_type": "str"}

    attribute_map = {"strategy_type": "strategyType", "linking_type": "linkingType"}

    discriminator_value_class_map = {
        "OPERATOR": "OperatorLinkingConfig",
        "ID_EQUALS": "IdEqualsLinkingConfig",
    }

    def __init__(
        self, strategy_type="LINKING", linking_type=None, local_vars_configuration=None
    ):  # noqa: E501
        """BaseLinkingConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._strategy_type = None
        self._linking_type = None
        self.discriminator = None

        self.strategy_type = strategy_type
        self.linking_type = linking_type

    @property
    def strategy_type(self):
        """Gets the strategy_type of this BaseLinkingConfig.  # noqa: E501


        :return: The strategy_type of this BaseLinkingConfig.  # noqa: E501
        :rtype: str
        """
        return self._strategy_type

    @strategy_type.setter
    def strategy_type(self, strategy_type):
        """Sets the strategy_type of this BaseLinkingConfig.


        :param strategy_type: The strategy_type of this BaseLinkingConfig.  # noqa: E501
        :type: str
        """
        if (
            self.local_vars_configuration.client_side_validation
            and strategy_type is None
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `strategy_type`, must not be `None`"
            )  # noqa: E501
        allowed_values = ["LINKING", "FUSING", "PREDICTING"]  # noqa: E501
        if (
            self.local_vars_configuration.client_side_validation
            and strategy_type not in allowed_values
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `strategy_type` ({0}), must be one of {1}".format(  # noqa: E501
                    strategy_type, allowed_values
                )
            )

        self._strategy_type = strategy_type

    @property
    def linking_type(self):
        """Gets the linking_type of this BaseLinkingConfig.  # noqa: E501


        :return: The linking_type of this BaseLinkingConfig.  # noqa: E501
        :rtype: str
        """
        return self._linking_type

    @linking_type.setter
    def linking_type(self, linking_type):
        """Sets the linking_type of this BaseLinkingConfig.


        :param linking_type: The linking_type of this BaseLinkingConfig.  # noqa: E501
        :type: str
        """
        if (
            self.local_vars_configuration.client_side_validation
            and linking_type is None
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `linking_type`, must not be `None`"
            )  # noqa: E501
        allowed_values = ["OPERATOR", "ID_EQUALS"]  # noqa: E501
        if (
            self.local_vars_configuration.client_side_validation
            and linking_type not in allowed_values
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `linking_type` ({0}), must be one of {1}".format(  # noqa: E501
                    linking_type, allowed_values
                )
            )

        self._linking_type = linking_type

    def get_real_child_model(self, data):
        """Returns the child model by discriminator"""
        if "@type" in data:
            child_type = data.get("@type")
            real_child_model = self.discriminator_value_class_map.get(child_type)
            return real_child_model
        return None

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, BaseLinkingConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, BaseLinkingConfig):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


"""
    knext

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from knext.common.rest.configuration import Configuration


class OperatorConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        "file_path": "str",
        "module_path": "int",
        "class_name": "str",
        "method": "str",
        "params": "object",
    }

    attribute_map = {
        "file_path": "filePath",
        "module_path": "modulePath",
        "class_name": "className",
        "method": "method",
        "params": "params",
    }

    def __init__(
        self,
        file_path=None,
        module_path=None,
        class_name=None,
        method=None,
        params=None,
        local_vars_configuration=None,
    ):  # noqa: E501
        """OperatorConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._file_path = None
        self._module_path = None
        self._class_name = None
        self._method = None
        self._params = None
        self.discriminator = None

        if file_path is not None:
            self.file_path = file_path
        if module_path is not None:
            self.module_path = module_path
        if class_name is not None:
            self.class_name = class_name
        if method is not None:
            self.method = method
        if params is not None:
            self.params = params

    @property
    def file_path(self):
        """Gets the file_path of this OperatorConfig.  # noqa: E501


        :return: The file_path of this OperatorConfig.  # noqa: E501
        :rtype: str
        """
        return self._file_path

    @file_path.setter
    def file_path(self, file_path):
        """Sets the file_path of this OperatorConfig.


        :param file_path: The file_path of this OperatorConfig.  # noqa: E501
        :type: str
        """

        self._file_path = file_path

    @property
    def module_path(self):
        """Gets the module_path of this OperatorConfig.  # noqa: E501


        :return: The module_path of this OperatorConfig.  # noqa: E501
        :rtype: int
        """
        return self._module_path

    @module_path.setter
    def module_path(self, module_path):
        """Sets the module_path of this OperatorConfig.


        :param module_path: The module_path of this OperatorConfig.  # noqa: E501
        :type: int
        """

        self._module_path = module_path

    @property
    def class_name(self):
        """Gets the class_name of this OperatorConfig.  # noqa: E501


        :return: The class_name of this OperatorConfig.  # noqa: E501
        :rtype: str
        """
        return self._class_name

    @class_name.setter
    def class_name(self, class_name):
        """Sets the class_name of this OperatorConfig.


        :param class_name: The class_name of this OperatorConfig.  # noqa: E501
        :type: str
        """

        self._class_name = class_name

    @property
    def method(self):
        """Gets the method of this OperatorConfig.  # noqa: E501


        :return: The method of this OperatorConfig.  # noqa: E501
        :rtype: str
        """
        return self._method

    @method.setter
    def method(self, method):
        """Sets the method of this OperatorConfig.


        :param method: The method of this OperatorConfig.  # noqa: E501
        :type: str
        """

        self._method = method

    @property
    def params(self):
        """Gets the params of this OperatorConfig.  # noqa: E501


        :return: The params of this OperatorConfig.  # noqa: E501
        :rtype: object
        """
        return self._params

    @params.setter
    def params(self, params):
        """Sets the params of this OperatorConfig.


        :param params: The params of this OperatorConfig.  # noqa: E501
        :type: object
        """

        self._params = params

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OperatorConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OperatorConfig):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from __future__ import absolute_import

from knext.builder.rest.models.pipeline.edge import Edge
from knext.builder.rest.models.pipeline.pipeline import Pipeline
from knext.builder.rest.models.pipeline.node import Node
from knext.builder.rest.models.pipeline.config.base_predicting_config import (
    BasePredictingConfig,
)
from knext.builder.rest.models.pipeline.config.spg_type_mapping_node_configs import (
    SpgTypeMappingNodeConfigs,
)
from knext.builder.rest.models.pipeline.config.base_node_config import BaseNodeConfig
from knext.builder.rest.models.pipeline.config.mapping_config import MappingConfig
from knext.builder.rest.models.pipeline.config.overwrite_fusing_config import (
    OverwriteFusingConfig,
)
from knext.builder.rest.models.pipeline.config.relation_mapping_node_config import (
    RelationMappingNodeConfig,
)
from knext.builder.rest.models.pipeline.config.operator_predicting_config import (
    OperatorPredictingConfig,
)
from knext.builder.rest.models.pipeline.config.base_strategy_config import (
    BaseStrategyConfig,
)
from knext.builder.rest.models.pipeline.config.id_equals_linking_config import (
    IdEqualsLinkingConfig,
)
from knext.builder.rest.models.pipeline.config.llm_based_extract_node_config import (
    LlmBasedExtractNodeConfig,
)
from knext.builder.rest.models.pipeline.config.base_fusing_config import (
    BaseFusingConfig,
)
from knext.builder.rest.models.pipeline.config.spg_type_mapping_node_config import (
    SpgTypeMappingNodeConfig,
)
from knext.builder.rest.models.pipeline.config.operator_config import OperatorConfig
from knext.builder.rest.models.pipeline.config.user_defined_extract_node_config import (
    UserDefinedExtractNodeConfig,
)
from knext.builder.rest.models.pipeline.config.graph_store_sink_node_config import (
    GraphStoreSinkNodeConfig,
)
from knext.builder.rest.models.pipeline.config.mapping_filter import MappingFilter
from knext.builder.rest.models.pipeline.config.csv_source_node_config import (
    CsvSourceNodeConfig,
)
from knext.builder.rest.models.pipeline.config.operator_linking_config import (
    OperatorLinkingConfig,
)
from knext.builder.rest.models.pipeline.config.predicting_config import PredictingConfig
from knext.builder.rest.models.pipeline.config.operator_fusing_config import (
    OperatorFusingConfig,
)
from knext.builder.rest.models.pipeline.config.base_linking_config import (
    BaseLinkingConfig,
)
from knext.builder.rest.models.response.builder_job_inst import BuilderJobInst
from knext.builder.rest.models.response.search_engine_index_response import (
    SearchEngineIndexResponse,
)
from knext.builder.rest.models.request.builder_job_submit_request import (
    BuilderJobSubmitRequest,
)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os
import string
from pathlib import Path

import click

from knext.schema.marklang.concept_rule_ml import SPGConceptRuleMarkLang
from knext.schema.marklang.schema_ml import SPGSchemaMarkLang
from knext.common.template import copytree

TEMPLATE_TO_RENDER = "${project}_schema_helper.py.tmpl"


def list_schema():
    """
    List all server-side schemas.
    """
    click.secho("Not support yes.", fg="bright_yellow")


def diff_schema():
    """
    Print differences of schema between local and server.
    """
    schema_path = os.path.join(
        os.environ["KNEXT_ROOT_PATH"],
        os.environ["KNEXT_SCHEMA_DIR"],
        os.environ["KNEXT_SCHEMA_FILE"],
    )
    ml = SPGSchemaMarkLang(schema_path)
    ml.print_diff()


def commit_schema():
    """
    Commit local schema and generate schema helper.
    """
    schema_file = os.path.join(
        os.environ["KNEXT_ROOT_PATH"],
        os.environ["KNEXT_SCHEMA_DIR"],
        os.environ["KNEXT_SCHEMA_FILE"],
    )
    if not Path(schema_file).exists():
        click.secho(f"ERROR: File {schema_file} not exists.", fg="bright_red")
        return

    ml = SPGSchemaMarkLang(schema_file)
    is_altered = ml.sync_schema()

    helper_file = os.path.join(
        os.environ["KNEXT_ROOT_PATH"],
        os.environ["KNEXT_SCHEMA_DIR"],
        TEMPLATE_TO_RENDER,
    )
    copytree(
        Path("schema_helper"), Path(helper_file).parent, os.environ["KNEXT_PROJECT_DIR"]
    )
    tplfile = string.Template(helper_file).substitute(
        project=os.environ["KNEXT_PROJECT_DIR"]
    )
    ml.export_schema_python(tplfile)
    if is_altered:
        click.secho("Schema is successfully committed.", fg="bright_green")
        click.secho(
            f"SchemaHelper is created in {os.environ['KNEXT_SCHEMA_DIR']}/{os.environ['KNEXT_PROJECT_DIR']}_schema_helper.py.",
            fg="bright_green",
        )
    else:
        click.secho(
            "There is no diff between local and server-side schema.", fg="bright_yellow"
        )


@click.option("--file", help="Path of DSL file.")
def reg_concept_rule(file):
    """
    Register a concept rule according to DSL file.
    """
    SPGConceptRuleMarkLang(file)
    click.secho(f"Concept rule is successfully registered", fg="bright_green")



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
from typing import Dict, List

from knext.builder.operator import ExtractOp
from knext.builder.operator.spg_record import SPGRecord
from nn4k.invoker import NNInvoker


class _BuiltInOnlineExtractor(ExtractOp):
    def __init__(self, params: Dict[str, str] = None):
        super().__init__(params)
        self.debug = bool(params.get("debug", ""))
        self.max_retry_times = int(self.params.get("max_retry_times", "3"))
        self.model = self.load_model()
        self.prompt_ops = self.load_operator()

    def load_model(self):
        model_config = json.loads(self.params["model_config"])
        return NNInvoker.from_config(model_config)

    def load_operator(self):
        import importlib.util

        prompt_config = json.loads(self.params["prompt_config"])
        prompt_ops = []
        for op_config in prompt_config:
            spec = importlib.util.spec_from_file_location(
                op_config["modulePath"], op_config["filePath"]
            )
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            op_clazz = getattr(module, op_config["className"])
            params = op_config.get("params", {})
            op_obj = op_clazz(*params.values())
            if self.debug:
                print(f'{op_config["className"]}.template: {op_obj.template}')
            prompt_ops.append(op_obj)

        return prompt_ops

    def invoke(self, record: Dict[str, str]) -> List[SPGRecord]:

        collector = []
        input_params = [record]
        for op in self.prompt_ops:
            next_params = []
            for input_param in input_params:
                retry_times = 0
                while retry_times < self.max_retry_times:
                    try:
                        querys = op.build_prompt(input_param)
                        if isinstance(querys, str):
                            querys = [querys]
                        for query in querys:
                            response = self.model.remote_inference(query)
                            collector.extend(op.parse_response(response))
                            next_params.extend(
                                op._build_next_variables(input_param, response)
                            )
                        break
                    except Exception as e:
                        retry_times += 1
                        print(e)
            input_params = next_params
        if self.debug:
            print(f"LLMBasedExtractor results: {collector}")
        return collector



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import pprint
from typing import Dict, List, Optional

from elasticsearch import Elasticsearch

from knext.builder import rest
from knext.schema.model.schema_helper import PropertyName
from knext.builder.operator.spg_record import SPGRecord


class IdxRecord:
    """Represents a record retrieved from an index.

    Attributes:
        index_name (str): The name of the index where the record belongs.
        doc_id (str): The unique identifier of the document.
        score (float): The relevance score of the record.
        properties (Dict[str, str]): A dictionary containing the properties or fields of the record.

    """

    def __init__(
        self,
        spg_type_name: str,
        index_name: str,
        doc_id: str,
        score: float,
        properties: Dict[str, str],
    ):
        self.spg_type_name = spg_type_name
        self.index_name = index_name
        self.doc_id = doc_id
        self.score = score
        self.properties = properties

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.__dict__)

    def to_spg_record(self):
        record = SPGRecord(self.spg_type_name)
        record.upsert_property("id", self.doc_id)
        for name, value in self.properties.items():
            record.upsert_property(name, value)
        return record


class SearchClient:
    """Client connected to search engine, which can be imported in operator to recall entities.
    You can initialize this client in `BaseOp.__init__()` and invoke `search` method in `BaseOp.eval()`.

    """

    def __init__(self, spg_type_name: str):
        _client = rest.BuilderApi()
        response = _client.search_engine_index_get(spg_type=spg_type_name)

        self.index_name = response.index_name
        self.spg_type_name = spg_type_name
        self.client = Elasticsearch("http://127.0.0.1:9200")

    def search(self, query, sort=None, filter=None, start: int = 0, size: int = 10):
        """Perform a search operation on the specified index using the given query.

        Args:
            query: The query to be executed for the search.
            sort: Optional. The sorting criteria for the search results.
            filter: Optional. The filter to be applied to the search results.
            start: Optional. The starting position of the search results to be returned. Default is 0.
            size: Optional. The maximum number of search results to be returned. Default is 10.

        Returns:
            A list of `IdxRecord` objects representing the search results, or None if no results were found.

        """
        data = self.client.search(
            index=self.index_name,
            query=query,
            sort=sort,
            post_filter=filter,
            from_=start,
            size=size,
        )
        if "hits" in data and "hits" in data.get("hits"):
            hits = data.get("hits").get("hits")
            records = []
            for hit in hits:
                records.append(
                    IdxRecord(
                        self.spg_type_name,
                        hit.get("_index"),
                        hit.get("_id"),
                        hit.get("_score"),
                        hit.get("_source"),
                    )
                )
            return records
        return None

    def fuzzy_search(
        self, record: SPGRecord, property_name: PropertyName, size: int = 10
    ) -> List[SPGRecord]:
        property_value = record.get_property(property_name)
        if not property_value:
            return []
        query = {"match": {property_name: property_value}}
        records = []
        recall_results = self.search(query, size=size)
        if recall_results:
            for recall_result in recall_results:
                records.append(recall_result.to_spg_record())
        return records

    def exact_search(
        self, record: SPGRecord, property_name: PropertyName
    ) -> Optional[SPGRecord]:
        property_value = record.get_property(property_name)
        if not property_value:
            return None
        query = {"match": {property_name: property_value}}
        recall_results = self.search(query, size=1)
        if recall_results:
            if recall_results[0].properties.get(property_name) == property_value:
                return recall_results[0].to_spg_record()
        return None

    def exact_search_by_property(
        self, property_value: str, property_name: PropertyName
    ) -> Optional[SPGRecord]:
        query = {"match": {property_name: property_value}}
        recall_results = self.search(query, size=1)
        if recall_results:
            if recall_results[0].properties.get(property_name) == property_value:
                return recall_results[0].to_spg_record()
        return None

    def fuzzy_search_by_property(
        self, property_value: str, property_name: PropertyName, size: int = 10
    ) -> List[SPGRecord]:
        query = {"match": {property_name: property_value}}
        records = []
        recall_results = self.search(query, size=size)
        if recall_results:
            for recall_result in recall_results:
                records.append(recall_result.to_spg_record())
        return records



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import re

from knext.schema import rest
from knext.schema.client import SchemaClient
from knext.schema.model.base import SpgTypeEnum


class SPGConceptRuleMarkLang:
    """
    SPG Concept Rule Mark Language Parser
    Feature: parse rule script and then alter the schema of project
    """

    namespace = None
    rule_quote_open = False
    rule_text = ""
    src_concept = ()
    dst_concept = ()

    def __init__(self, filename):
        self.current_line_num = 0
        self.session = SchemaClient().create_session()
        self.concept_client = rest.ConceptApi()
        self.load_script(filename)

    def error_msg(self, msg):
        return f"Line# {self.current_line_num}: {msg}"

    def parse_concept(self, expression):
        """
        parse the concept definition
        """

        namespace_match = re.match(r"^namespace\s+([a-zA-Z0-9]+)$", expression)
        if namespace_match:
            assert self.namespace is None, self.error_msg(
                "Duplicated namespace define, please ensure define it only once"
            )

            self.namespace = namespace_match.group(1)
            return

        type_match = re.match(
            r"^`([a-zA-Z0-9\.]+)`/`([^`]+)`:(\s*?([a-zA-Z0-9\.]+)/`([^`]+)`)?$",
            expression,
        )
        if type_match:
            assert self.namespace is not None, self.error_msg(
                "please define namespace first"
            )

            self.src_concept = (type_match.group(1), type_match.group(2))
            if len(type_match.groups()) > 4:
                self.dst_concept = (type_match.group(4), type_match.group(5))

        else:
            raise Exception(
                self.error_msg("parse error, expect `ConceptType`/`ConceptName`:")
            )

    def parse_rule(self, rule):
        """
        parse the logic rule from text
        """

        strip_rule = rule.strip()
        if strip_rule.startswith("[["):
            self.rule_quote_open = True
            if len(strip_rule) > 2:
                if strip_rule.endswith("]]"):
                    self.rule_quote_open = False
                    self.rule_text = strip_rule[2 : len(strip_rule) - 2].lstrip()
                else:
                    self.rule_text = strip_rule[2].lstrip()
            else:
                self.rule_text = ""
        else:
            self.rule_text = rule

    def complete_rule(self, rule):
        """
        Auto generate define statement and append namespace to the entity name
        """

        pattern = re.compile(r"Define\s*\(", re.IGNORECASE)
        match = pattern.match(rule.strip())
        if not match:
            subject_type = None
            subject_name = None
            if self.dst_concept[0] is not None:
                predicate_name = "leadTo"
                subject_type = f"{self.namespace}.{self.src_concept[0]}"
                subject_name = self.src_concept[1]
                object_type = f"{self.namespace}.{self.dst_concept[0]}"
                object_name = self.dst_concept[1]
            else:
                predicate_name = "belongTo"
                object_type = f"{self.namespace}.{self.src_concept[0]}"
                object_name = self.src_concept[1]
                assert object_type in self.session.spg_types, self.error_msg(
                    f"{object_type} not found in schema"
                )

                concept_type = self.session.get(object_type)
                assert (
                    concept_type.spg_type_enum == SpgTypeEnum.Concept
                ), self.error_msg(f"{object_type} is not concept type")

                for spg_type in self.session.spg_types.values():
                    for relation_name in spg_type.relations:
                        if relation_name.startswith(f"belongTo_{object_type}"):
                            subject_type = spg_type.name
                            break

            if subject_name is None:
                head = (
                    f"Define (s:{subject_type})-[p:{predicate_name}]->(o:`{object_type}`/`{object_name}`)"
                    + " {\n"
                )
            else:
                head = (
                    f"Define "
                    f"(s:`{subject_type}`/`{subject_name}`)-[p:{predicate_name}]->(o:`{object_type}`/`{object_name}`)"
                    + " {\n"
                )
            rule = head + rule
            rule += "\n}"

        # complete the namespace of concept type
        pattern = re.compile(r"\(([\w\s]*?:)`([\w\s\.]+)`/`([^`]+)`\)", re.IGNORECASE)
        replace_list = []
        matches = re.findall(pattern, rule)
        if matches:
            for group in matches:
                if "." in group[1]:
                    continue
                replace_list.append(
                    (
                        f"({group[0]}`{group[1]}`",
                        f"({group[0]}`{self.namespace}.{group[1].strip()}`",
                    )
                )

        # complete the namespace of non-concept type
        pattern = re.compile(r"\(([\w\s]*?:)([\w\s\.]+)\)", re.IGNORECASE)
        matches = re.findall(pattern, rule)
        if matches:
            for group in matches:
                if "." not in group[1]:
                    replace_list.append(
                        (
                            f"({group[0]}{group[1]})",
                            f"({group[0]}{self.namespace}.{group[1].strip()})",
                        )
                    )

        # complete the namespace of type in action clause
        pattern = re.compile(
            r"createNodeInstance\s*?\([^)]+(type=)([^,]+),", re.IGNORECASE
        )
        matches = re.findall(pattern, rule)
        if matches:
            for group in matches:
                if "." not in group[1]:
                    replace_list.append(
                        (
                            f"{group[0]}{group[1]}",
                            f"{group[0]}{self.namespace}.{group[1].strip()}",
                        )
                    )

        if len(replace_list) > 0:
            for t in replace_list:
                rule = rule.replace(t[0], t[1])

        return rule

    def clear_session(self):
        self.src_concept = ()
        self.dst_concept = ()
        self.rule_text = ""

    def submit_rule(self):
        """
        submit the rule definition, make them available for inference
        """

        if self.dst_concept[0] is None:
            # belongTo rule
            self.concept_client.concept_define_dynamic_taxonomy_post(
                define_dynamic_taxonomy_request=rest.DefineDynamicTaxonomyRequest(
                    concept_type_name=f"{self.namespace}.{self.src_concept[0]}",
                    concept_name=self.src_concept[1],
                    dsl=self.rule_text,
                )
            )
            print(
                f"Defined belongTo rule for `{self.src_concept[0]}`/`{self.src_concept[1]}`"
            )

        else:
            # leadTo rule
            self.concept_client.concept_define_logical_causation_post(
                define_logical_causation_request=rest.DefineLogicalCausationRequest(
                    subject_concept_type_name=f"{self.namespace}.{self.src_concept[0]}",
                    subject_concept_name=self.src_concept[1],
                    predicate_name="leadTo",
                    object_concept_type_name=f"{self.namespace}.{self.dst_concept[0]}",
                    object_concept_name=self.dst_concept[1],
                    dsl=self.rule_text,
                )
            )
            print(
                f"Defined leadTo rule for "
                f"`{self.src_concept[0]}`/`{self.src_concept[1]}` -> `{self.dst_concept[0]}`/`{self.dst_concept[1]}`"
            )
        self.clear_session()

    def load_script(self, filename):
        """
        Load and then parse the script file
        """

        file = open(filename, "r", encoding="utf-8")
        lines = file.read().splitlines()
        last_indent_level = 0

        for line in lines:
            self.current_line_num += 1
            strip_line = line.strip()
            if strip_line == "" or strip_line.startswith("#"):
                # skip empty or comments line
                continue

            if self.rule_quote_open:
                # process the multi-line assignment [[ .... ]]
                right_strip_line = line.rstrip()
                if strip_line.endswith("]]"):
                    self.rule_quote_open = False
                    if len(right_strip_line) > 2:
                        self.rule_text += right_strip_line[: len(right_strip_line) - 2]
                    self.rule_text = self.complete_rule(self.rule_text)
                    self.submit_rule()

                else:
                    self.rule_text += line + "\n"
                continue
            elif len(self.rule_text) > 0:
                self.submit_rule()

            indent_count = len(line) - len(line.lstrip())
            if indent_count == 0:
                # the line without indent is namespace definition or a concept definition
                self.clear_session()
                self.parse_concept(strip_line)

            elif indent_count > last_indent_level:
                # the line is the sub definition of the previous line
                if strip_line.startswith("rule:"):
                    if len(strip_line) > 5:
                        self.parse_rule(strip_line[5:])
                else:
                    raise Exception(self.error_msg("parse error, expect rule:"))

            last_indent_level = indent_count

        # if rule is the last line of file, then submit it
        if len(self.rule_text) > 0:
            self.submit_rule()



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from cachetools import TTLCache


class LinkCache:
    def __init__(self, maxsize: int = 500, ttl: int = 60):
        self._cache = TTLCache(maxsize=maxsize, ttl=ttl)

    @property
    def cache(self):
        return self._cache

    def put(self, key, value):
        self.cache[key] = value

    def get(self, key):
        return self.cache.get(key)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from typing import List, Type, Union, Dict

from knext.schema.model.base import (
    ConstraintTypeEnum,
    PropertyGroupEnum,
    BaseProperty,
)


class Property(BaseProperty):
    """Property Model."""

    name: str
    object_type_name: str
    name_zh: str
    desc: str
    property_group: PropertyGroupEnum
    sub_properties: Dict[str, Type["Property"]]
    constraint: Dict[ConstraintTypeEnum, Union[str, List[str]]]
    logical_rule: str

    def __init__(
        self,
        name: str,
        object_type_name: str,
        name_zh: str = None,
        desc: str = None,
        property_group: PropertyGroupEnum = None,
        sub_properties: List[Type["Property"]] = None,
        constraint: Dict[ConstraintTypeEnum, Union[str, List[str]]] = None,
        logical_rule: str = None,
        **kwargs
    ):
        super().__init__(
            name=name,
            object_type_name=object_type_name,
            name_zh=name_zh,
            desc=desc,
            property_group=property_group,
            sub_properties=sub_properties,
            constraint=constraint,
            logical_rule=logical_rule,
            **kwargs
        )



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
from knext.builder import rest
from knext.common.base.chain import Chain


class BuilderChain(Chain):
    @property
    def input_types(self):
        return None

    @property
    def output_types(self):
        return None

    def invoke(self, **kwargs):
        from knext.builder.client import BuilderClient

        client = BuilderClient()
        client.execute(self, **kwargs)

    @classmethod
    def from_chain(cls, chain):
        return cls(dag=chain.dag)

    def to_rest(self) -> rest.Pipeline:
        nodes, edges = [], []
        for node in self.dag.nodes:
            nodes.append(node.to_rest())
        for edge in self.dag.edges:
            edges.append(rest.Edge(_from=edge[0].id, to=edge[1].id))
        dag_config = rest.Pipeline(nodes=nodes, edges=edges)
        return dag_config



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import re
from enum import Enum
from pathlib import Path

from knext.schema.model.base import (
    HypernymPredicateEnum,
    BasicTypeEnum,
    ConstraintTypeEnum,
    AlterOperationEnum,
    SpgTypeEnum,
    PropertyGroupEnum,
)
from knext.schema.model.spg_type import (
    EntityType,
    ConceptType,
    EventType,
    StandardType,
    Property,
    Relation,
)
from knext.schema.client import SchemaClient


class IndentLevel(Enum):
    # Define entity/concept/event/standard types or subtypes
    Type = 0

    # Define description/properties/relations of type
    TypeMeta = 1

    # Define property/relation name of type
    Predicate = 2

    # Define description/constraint/rule of property/relation
    PredicateMeta = 3

    # Define property about property
    SubProperty = 4

    # Define constraint of sub property
    SubPropertyMeta = 5


class RegisterUnit(Enum):
    Type = "type"
    Property = "property"
    Relation = "relation"
    SubProperty = "subProperty"


class SPGSchemaMarkLang:
    """
    SPG Schema Mark Language Parser
    Feature1: parse schema script and then alter the schema of project
    Feature2: export schema script from a project
    """

    internal_type = set()
    entity_internal_property = set()
    event_internal_property = {"eventTime"}
    concept_internal_property = {"stdId", "alias"}
    keyword_type = {"EntityType", "ConceptType", "EventType", "StandardType"}
    semantic_rel = {
        "SYNANT": [
            "synonym",
            "antonym",
            "symbolOf",
            "distinctFrom",
            "definedAs",
            "locatedNear",
            "similarTo",
            "etymologicallyRelatedTo",
        ],
        "CAU": ["leadTo", "causes", "obstructedBy", "createdBy", "causesDesire"],
        "SEQ": [
            "happenedBefore",
            "hasSubevent",
            "hasFirstSubevent",
            "hasLastSubevent",
            "hasPrerequisite",
        ],
        "IND": ["belongTo"],
        "INC": ["isPartOf", "hasA", "madeOf", "derivedFrom", "hasContext"],
        "USE": ["usedFor", "capableOf", "receivesAction", "motivatedByGoal"],
    }
    semantic_rel_zh = {
        "synonym": "同义",
        "antonym": "反义",
        "symbolOf": "象征",
        "distinctFrom": "区别于",
        "definedAs": "定义为",
        "locatedNear": "位置临近",
        "similarTo": "类似于",
        "etymologicallyRelatedTo": "词源相关",
        "leadTo": "导致",
        "causes": "引起",
        "obstructedBy": "受阻于",
        "createdBy": "由...创建",
        "causesDesire": "引发欲望",
        "happenedBefore": "先于...发生",
        "hasSubevent": "拥有子事件",
        "hasFirstSubevent": "以...作为开始",
        "hasLastSubevent": "以...作为结束",
        "hasPrerequisite": "有前提条件",
        "belongTo": "属于",
        "isPartOf": "是...的一部分",
        "hasA": "拥有",
        "madeOf": "由…制成",
        "derivedFrom": "源自于",
        "hasContext": "有上下文",
        "usedFor": "用于",
        "capableOf": "能够",
        "receivesAction": "接受动作",
        "motivatedByGoal": "目标驱动",
    }
    parsing_register = {
        RegisterUnit.Type: None,
        RegisterUnit.Property: None,
        RegisterUnit.Relation: None,
        RegisterUnit.SubProperty: None,
    }
    indent_level_pos = [None, None, None, None, None, None]
    rule_quote_predicate = None
    rule_quote_open = False
    current_parsing_level = 0
    last_indent_level = 0
    namespace = None
    types = {}
    defined_types = {}

    def __init__(self, filename):
        self.reset()
        self.schema_file = filename
        self.current_line_num = 0
        self.schema = SchemaClient()
        thing = self.schema.query_spg_type("Thing")
        for prop in thing.properties:
            self.entity_internal_property.add(prop)
            self.event_internal_property.add(prop)
            self.concept_internal_property.add(prop)
        session = self.schema.create_session()
        for type_name in session.spg_types:
            spg_type = session.get(type_name)
            if session.get(type_name).spg_type_enum in [
                SpgTypeEnum.Basic,
                SpgTypeEnum.Standard,
            ]:
                self.internal_type.add(spg_type.name)
        self.load_script()

    def reset(self):
        self.internal_type = set()
        self.entity_internal_property = set()
        self.event_internal_property = {"eventTime"}
        self.concept_internal_property = {"stdId", "alias"}
        self.keyword_type = {"EntityType", "ConceptType", "EventType", "StandardType"}

        self.parsing_register = {
            RegisterUnit.Type: None,
            RegisterUnit.Property: None,
            RegisterUnit.Relation: None,
            RegisterUnit.SubProperty: None,
        }
        self.indent_level_pos = [None, None, None, None, None, None]
        self.rule_quote_predicate = None
        self.rule_quote_open = False
        self.current_parsing_level = 0
        self.last_indent_level = 0
        self.namespace = None
        self.types = {}

    def save_register(self, element: RegisterUnit, value):
        """
        maintain the session for parsing
        """

        self.parsing_register[element] = value
        if element == RegisterUnit.Type:
            self.parsing_register[RegisterUnit.Property] = None
            self.parsing_register[RegisterUnit.Relation] = None
            self.parsing_register[RegisterUnit.SubProperty] = None
        elif element == RegisterUnit.Property:
            self.parsing_register[RegisterUnit.Relation] = None
            self.parsing_register[RegisterUnit.SubProperty] = None
        elif element == RegisterUnit.Relation:
            self.parsing_register[RegisterUnit.Property] = None
            self.parsing_register[RegisterUnit.SubProperty] = None

    def adjust_parsing_level(self, step):
        """
        mark the indent level and clear related session
        """

        if step == 0:
            self.current_parsing_level = IndentLevel.Type.value
            # finish type parsing, clear the field in session
            self.save_register(RegisterUnit.Type, None)
            return
        if step < 0:
            self.current_parsing_level = max(0, self.current_parsing_level + step)
            if self.current_parsing_level == IndentLevel.PredicateMeta.value:
                # finish sub property parsing, clear the field in session
                self.save_register(RegisterUnit.SubProperty, None)
            elif self.current_parsing_level == IndentLevel.Predicate.value:
                # finish predicate parsing, clear the fields in session
                if self.parsing_register[RegisterUnit.Property] is not None:
                    self.save_register(
                        RegisterUnit.Property,
                        Property(name="_", object_type_name="Thing"),
                    )
                elif self.parsing_register[RegisterUnit.Relation] is not None:
                    self.save_register(
                        RegisterUnit.Relation,
                        Relation(name="_", object_type_name="Thing"),
                    )
        elif step == 1:
            assert self.current_parsing_level + 1 < len(IndentLevel), self.error_msg(
                "Invalid indentation (too many levels?)"
            )

            self.current_parsing_level += 1

    def error_msg(self, msg):
        return f"Line# {self.current_line_num}: {msg} (Please refer https://spg.openkg.cn/tutorial/schema/dsl for details)"

    def get_type_name_with_ns(self, type_name: str):
        if "." in type_name:
            return type_name
        else:
            return f"{self.namespace}.{type_name}"

    def parse_type(self, expression):
        """
        parse the SPG type definition
        """

        namespace_match = re.match(r"^namespace\s+([a-zA-Z0-9]+)$", expression)
        if namespace_match:
            assert self.namespace is None, self.error_msg(
                "Duplicated namespace define, please ensure define it only once"
            )

            self.namespace = namespace_match.group(1)
            return

        type_match = re.match(
            r"^([a-zA-Z0-9\.]+)\((\w+)\):\s*?([a-zA-Z0-9,]+)$", expression
        )
        if type_match:
            assert self.namespace is not None, self.error_msg(
                "Missing namespace, please define namespace at the first"
            )

            type_name = type_match.group(1)
            type_name_zh = type_match.group(2).strip()
            type_class = type_match.group(3).strip()
            assert type_class in self.keyword_type, self.error_msg(
                f"{type_class} is illegal, please define it before current line"
            )
            assert (
                type_name.startswith("STD.")
                or "." not in type_name
                or type_name.startswith(f"{self.namespace}.")
            ), self.error_msg(
                f"The name space of {type_name} does not belong to current project."
            )

            spg_type = None
            if type_class == "EntityType":
                spg_type = EntityType(
                    name=self.get_type_name_with_ns(type_name), name_zh=type_name_zh
                )
            elif type_class == "ConceptType":
                spg_type = ConceptType(
                    name=self.get_type_name_with_ns(type_name),
                    name_zh=type_name_zh,
                    hypernym_predicate=HypernymPredicateEnum.IsA,
                )
            elif type_class == "EventType":
                spg_type = EventType(
                    name=self.get_type_name_with_ns(type_name), name_zh=type_name_zh
                )
            elif type_class == "StandardType":
                spg_type = StandardType(name=f"{type_name}", name_zh=type_name_zh)
                spg_type.spreadable = False
                assert type_name.startswith("STD."), self.error_msg(
                    "The name of standard type must start with STD."
                )
            ns_type_name = self.get_type_name_with_ns(type_name)
            assert ns_type_name not in self.types, self.error_msg(
                f'Type "{type_name}" is duplicated in the schema'
            )

            self.types[ns_type_name] = spg_type
            self.save_register(RegisterUnit.Type, spg_type)
            return

        sub_type_match = re.match(
            r"^([a-zA-Z0-9]+)\((\w+)\)\s*?->\s*?([a-zA-Z0-9\.]+):$", expression
        )
        if sub_type_match:
            assert self.namespace is not None, self.error_msg(
                "Missing namespace, please define namespace at the first"
            )

            type_name = sub_type_match.group(1)
            type_name_zh = sub_type_match.group(2).strip()
            type_class = sub_type_match.group(3).strip()
            if "." not in type_class:
                ns_type_class = self.get_type_name_with_ns(type_class)
            else:
                ns_type_class = type_class
            assert (
                type_class not in self.keyword_type
                and type_class not in self.internal_type
            ), self.error_msg(f"{type_class} is not a valid inheritable type")
            assert ns_type_class in self.types, self.error_msg(
                f"{type_class} not found, please define it first"
            )

            parent_spg_type = self.types[ns_type_class]
            assert parent_spg_type.spg_type_enum in [
                SpgTypeEnum.Entity,
                SpgTypeEnum.Event,
            ], self.error_msg(
                f'"{type_class}" cannot be inherited, only entity/event type can be inherited.'
            )

            spg_type = EntityType(
                name=f"{self.namespace}.{type_name}", name_zh=type_name_zh
            )
            if parent_spg_type.spg_type_enum == SpgTypeEnum.Event:
                spg_type = EventType(
                    name=f"{self.namespace}.{type_name}", name_zh=type_name_zh
                )
            spg_type.name = type_name
            spg_type.name_zh = type_name_zh
            spg_type.parent_type_name = ns_type_class
            ns_type_name = f"{self.namespace}.{type_name}"
            self.types[ns_type_name] = spg_type
            self.save_register(RegisterUnit.Type, spg_type)
            return

        raise Exception(
            self.error_msg(
                "unrecognized expression, expect namespace A or A(B):C or A(B)->C"
            )
        )

    def parse_type_meta(self, expression):
        """
        parse the meta definition of SPG type
        """

        match = re.match(
            r"^(desc|properties|relations|hypernymPredicate|regular|spreadable|autoRelate):\s*?(.*)$",
            expression,
        )
        assert match, self.error_msg(
            "Unrecognized expression, expect desc:|properties:|relations:"
        )

        type_meta = match.group(1)
        meta_value = match.group(2).strip()

        if type_meta == "desc" and len(meta_value) > 0:
            self.parsing_register[RegisterUnit.Type].desc = meta_value

        elif type_meta == "properties":
            assert self.parsing_register[RegisterUnit.Type].spg_type_enum not in [
                SpgTypeEnum.Standard,
                SpgTypeEnum.Concept,
            ], self.error_msg(
                "Standard/concept type does not allow defining properties."
            )
            self.save_register(
                RegisterUnit.Property, Property(name="_", object_type_name="Thing")
            )

        elif type_meta == "relations":
            assert self.parsing_register[RegisterUnit.Type].spg_type_enum not in [
                SpgTypeEnum.Standard
            ], self.error_msg("Standard type does not allow defining relations.")
            self.save_register(
                RegisterUnit.Relation, Relation(name="_", object_type_name="Thing")
            )

        elif type_meta == "hypernymPredicate":
            assert meta_value in ["isA", "locateAt", "mannerOf"], self.error_msg(
                "Invalid hypernym predicate, expect isA or locateAt or mannerOf"
            )
            assert (
                self.parsing_register[RegisterUnit.Type].spg_type_enum
                == SpgTypeEnum.Concept
            ), self.error_msg("Hypernym predicate is available for concept type only")

            if meta_value == "isA":
                self.parsing_register[
                    RegisterUnit.Type
                ].hypernym_predicate = HypernymPredicateEnum.IsA
            elif meta_value == "locateAt":
                self.parsing_register[
                    RegisterUnit.Type
                ].hypernym_predicate = HypernymPredicateEnum.LocateAt
            elif meta_value == "mannerOf":
                self.parsing_register[
                    RegisterUnit.Type
                ].hypernym_predicate = HypernymPredicateEnum.MannerOf

        elif type_meta == "regular":
            assert (
                self.parsing_register[RegisterUnit.Type].spg_type_enum
                == SpgTypeEnum.Standard
            ), self.error_msg("Regular is available for standard type only")
            self.parsing_register[RegisterUnit.Type].constraint = {
                "REGULAR": meta_value
            }

        elif type_meta == "spreadable":
            assert (
                self.parsing_register[RegisterUnit.Type].spg_type_enum
                == SpgTypeEnum.Standard
            ), self.error_msg("Spreadable is available for standard type only")
            assert meta_value == "True" or meta_value == "False", self.error_msg(
                "Spreadable only accept True or False as its value"
            )
            self.parsing_register[RegisterUnit.Type].spreadable = meta_value == "True"

        elif type_meta == "autoRelate":
            assert (
                self.parsing_register[RegisterUnit.Type].spg_type_enum
                == SpgTypeEnum.Concept
            ), self.error_msg(
                "AutoRelate definition is available for concept type only"
            )
            concept_types = meta_value.split(",")
            for concept in concept_types:
                c = self.get_type_name_with_ns(concept.strip())
                assert (
                    c in self.types
                    and self.types[c].spg_type_enum == SpgTypeEnum.Concept
                ), self.error_msg(
                    f"{concept.strip()} is not a concept type, "
                    f"concept type only allow relationships defined between concept types"
                )
                for k in self.semantic_rel:
                    if k == "IND":
                        continue
                    for p in self.semantic_rel[k]:
                        predicate = Relation(
                            name=p, name_zh=self.semantic_rel_zh[p], object_type_name=c
                        )
                        self.parsing_register[RegisterUnit.Type].add_relation(predicate)
        return

    def check_semantic_relation(self, predicate_name, predicate_class):
        """
        Check if the definition of semantic relations is correct
        """

        name_arr = predicate_name.split("#")
        short_name = name_arr[0]
        pred_name = name_arr[1]
        assert short_name in self.semantic_rel, self.error_msg(
            f"{short_name} is incorrect, expect SYNANT/CAU/SEQ/IND/INC"
        )
        assert pred_name in self.semantic_rel[short_name], self.error_msg(
            f'{pred_name} is incorrect, expect {" / ".join(self.semantic_rel[short_name])}'
        )

        subject_type = self.parsing_register[RegisterUnit.Type]
        predicate_class_ns = predicate_class
        if "." not in predicate_class:
            predicate_class_ns = f"{self.namespace}.{predicate_class}"
        assert (
            predicate_class_ns in self.types or predicate_class_ns in self.defined_types
        ), self.error_msg(
            f"{predicate_class} is illegal, please ensure that it appears in this schema."
        )
        object_type = self.types[predicate_class_ns]

        if short_name == "SYNANT":
            assert subject_type.spg_type_enum == SpgTypeEnum.Concept, self.error_msg(
                "Only concept types could define synonym/antonym relation"
            )
            assert object_type.spg_type_enum == SpgTypeEnum.Concept, self.error_msg(
                "Synonymy/antonym relation can only point to concept types"
            )
        elif short_name == "CAU":
            assert subject_type.spg_type_enum in [
                SpgTypeEnum.Concept,
                SpgTypeEnum.Event,
            ], self.error_msg("Only concept/event types could define causal relation")
            assert object_type.spg_type_enum in [
                SpgTypeEnum.Concept,
                SpgTypeEnum.Event,
            ], self.error_msg(
                f'"{predicate_class}" must be a concept type to conform to the definition of causal relation'
            )
            if subject_type.spg_type_enum == SpgTypeEnum.Concept:
                assert object_type.spg_type_enum == SpgTypeEnum.Concept, self.error_msg(
                    "The causal relation of concept types can only point to concept types"
                )
        elif short_name == "SEQ":
            assert subject_type.spg_type_enum in [
                SpgTypeEnum.Event,
                SpgTypeEnum.Concept,
            ], self.error_msg(
                "Only concept/event types could define sequential relation"
            )
            assert (
                subject_type.spg_type_enum == object_type.spg_type_enum
            ), self.error_msg(
                f'"{predicate_class}" should keep the same type with "{subject_type.name.split(".")[1]}"'
            )
        elif short_name == "IND":
            assert subject_type.spg_type_enum in [
                SpgTypeEnum.Entity,
                SpgTypeEnum.Event,
            ], self.error_msg("Only entity/event types could define inductive relation")
            assert object_type.spg_type_enum == SpgTypeEnum.Concept, self.error_msg(
                f'"{predicate_class}" must be a concept type to conform to the definition of inductive relation'
            )
        elif short_name == "INC":
            assert subject_type.spg_type_enum == SpgTypeEnum.Concept, self.error_msg(
                "Only concept types could define inclusive relation"
            )
            assert object_type.spg_type_enum == SpgTypeEnum.Concept, self.error_msg(
                "The inclusion relation of concept types can only point to concept types"
            )
        elif short_name == "USE":
            assert subject_type.spg_type_enum == SpgTypeEnum.Concept, self.error_msg(
                "Only concept types could define usage relation"
            )
            assert object_type.spg_type_enum == SpgTypeEnum.Concept, self.error_msg(
                "The usage relation of concept types can only point to concept types"
            )

    def parse_predicate(self, expression):
        """
        parse the property/relation definition of SPG type
        """

        match = re.match(
            r"^([a-zA-Z0-9#]+)\(([\w\.]+)\):\s*?([a-zA-Z0-9,\.]+)$", expression
        )
        assert match, self.error_msg(
            "Unrecognized expression, expect pattern like english(Chinese):Type"
        )

        predicate_name = match.group(1)
        predicate_name_zh = match.group(2).strip()
        predicate_class = match.group(3).strip()
        cur_type = self.parsing_register[RegisterUnit.Type]
        type_name = cur_type.name

        if (
            cur_type.spg_type_enum == SpgTypeEnum.Concept
            and self.parsing_register[RegisterUnit.Relation] is None
        ):
            assert "#" in predicate_name, self.error_msg(
                "Concept type only accept following categories of relation: INC#/CAU#/SYNANT#/IND#/USE#/SEQ#"
            )

        if "#" in predicate_name:
            self.check_semantic_relation(predicate_name, predicate_class)
            predicate_name = predicate_name.split("#")[1]
        else:
            for semantic_short in self.semantic_rel.values():
                assert predicate_name not in semantic_short, self.error_msg(
                    f"{predicate_name} is a semantic predicate, please add the semantic prefix"
                )

        if (
            "." in predicate_class
            and predicate_class not in self.types
            and predicate_class not in self.internal_type
        ):
            try:
                cross_type = self.schema.query_spg_type(
                    self.get_type_name_with_ns(predicate_class)
                )
                self.types[self.get_type_name_with_ns(predicate_class)] = cross_type
            except Exception as e:
                raise ValueError(
                    self.error_msg(
                        f"{predicate_class} is illegal, please ensure the name space or type name is correct."
                    )
                )

        assert (
            self.get_type_name_with_ns(predicate_class) in self.types
            or predicate_class in self.internal_type
            or predicate_class in self.defined_types
        ), self.error_msg(
            f"{predicate_class} is illegal, please ensure that it appears in this schema."
        )

        assert predicate_name not in self.entity_internal_property, self.error_msg(
            f"property {predicate_name} is the default property of type"
        )
        if predicate_class not in self.internal_type:
            spg_type_enum = SpgTypeEnum.Entity
            if self.get_type_name_with_ns(predicate_class) in self.types:
                predicate_type = self.types[self.get_type_name_with_ns(predicate_class)]
                spg_type_enum = predicate_type.spg_type_enum
            elif predicate_class in self.defined_types:
                spg_type_enum_txt = self.defined_types[predicate_class]
                if spg_type_enum_txt == "EntityType":
                    spg_type_enum = SpgTypeEnum.Entity
                elif spg_type_enum_txt == "ConceptType":
                    spg_type_enum = SpgTypeEnum.Concept
                elif spg_type_enum_txt == "EventType":
                    spg_type_enum = SpgTypeEnum.Event
                elif spg_type_enum_txt == "StandardType":
                    spg_type_enum = SpgTypeEnum.Standard

            if cur_type.spg_type_enum == SpgTypeEnum.Concept:
                assert spg_type_enum == SpgTypeEnum.Concept, self.error_msg(
                    "Concept type only allow relationships that point to themselves"
                )
            elif cur_type.spg_type_enum == SpgTypeEnum.Entity:
                assert spg_type_enum != SpgTypeEnum.Event, self.error_msg(
                    "Relationships of entity types are not allowed to point to event types; "
                    "instead, they are only permitted to point from event types to entity types, "
                    "adhering to the principle of moving from dynamic to static."
                )

        if self.parsing_register[RegisterUnit.Relation] is not None:
            assert (
                predicate_name
                not in self.parsing_register[RegisterUnit.Relation].sub_properties
            ), self.error_msg(
                f'Property "{predicate_name}" is duplicated under the relation '
                f"{self.parsing_register[RegisterUnit.Relation].name}"
            )
        else:
            assert (
                predicate_name
                not in self.parsing_register[RegisterUnit.Type].properties
            ), self.error_msg(
                f'Property "{predicate_name}" is duplicated under the type {type_name[type_name.index(".") + 1:]}'
            )
        if predicate_class == "ConceptType":
            assert not self.is_internal_property(
                predicate_name, SpgTypeEnum.Concept
            ), self.error_msg(
                f"property {predicate_name} is the default property of ConceptType"
            )
        if predicate_class == "EventType":
            assert not self.is_internal_property(
                predicate_name, SpgTypeEnum.Event
            ), self.error_msg(
                f"property {predicate_name} is the default property of EventType"
            )

        if (
            "." not in predicate_class
            and predicate_class not in BasicTypeEnum.__members__
        ):
            predicate_class = f"{self.namespace}.{predicate_class}"

        if self.parsing_register[RegisterUnit.SubProperty]:
            # predicate is sub property
            predicate = Property(
                name=predicate_name,
                name_zh=predicate_name_zh,
                object_type_name=predicate_class,
            )
            if self.parsing_register[RegisterUnit.Property] is not None:
                self.parsing_register[RegisterUnit.Property].add_sub_property(predicate)
            elif self.parsing_register[RegisterUnit.Relation] is not None:
                self.parsing_register[RegisterUnit.Relation].add_sub_property(predicate)
            self.save_register(RegisterUnit.SubProperty, predicate)

        elif self.parsing_register[RegisterUnit.Property]:
            # predicate is property
            predicate = Property(
                name=predicate_name,
                name_zh=predicate_name_zh,
                object_type_name=predicate_class,
            )
            if predicate_class in self.types:
                predicate.object_spg_type = self.types[predicate_class].spg_type_enum
            if (
                self.parsing_register[RegisterUnit.Type].spg_type_enum
                == SpgTypeEnum.Event
                and predicate_name == "subject"
            ):
                assert predicate_class not in self.internal_type, self.error_msg(
                    f"The subject of event type only allows entity/concept type"
                )

                predicate.property_group = PropertyGroupEnum.Subject
                if "," in predicate_class:
                    # multi-types for subject
                    predicate.object_type_name = "Text"
                    subject_types = predicate_class.split(",")
                    for subject_type in subject_types:
                        subject_type = subject_type.strip()
                        assert (
                            subject_type not in BasicTypeEnum.__members__
                        ), self.error_msg(
                            f"{predicate_class} is illegal for subject in event type"
                        )

                        if "." not in subject_type:
                            subject_type = f"{self.namespace}.{predicate_class}"
                        assert (
                            subject_type in self.types
                            or predicate_class in self.defined_types
                        ), self.error_msg(
                            f"{predicate_class} is illegal, please ensure that it appears in this schema."
                        )

                        subject_predicate = Property(
                            name=f"subject{subject_type}",
                            name_zh=predicate_name_zh,
                            object_type_name=subject_type,
                        )
                        subject_predicate.property_group = PropertyGroupEnum.Subject
                        self.parsing_register[RegisterUnit.Type].add_property(
                            subject_predicate
                        )

            self.parsing_register[RegisterUnit.Type].add_property(predicate)
            self.save_register(RegisterUnit.Property, predicate)

        else:
            # predicate is relation
            assert not predicate_class.startswith("STD."), self.error_msg(
                f"{predicate_class} is not allow appear in the definition of relation."
            )
            assert (
                predicate_class in self.types
                or predicate_class.split(".")[1] in self.defined_types
            ), self.error_msg(
                f"{predicate_class} is illegal, please ensure that it appears in this schema."
            )
            assert (
                f"{predicate_name}_{predicate_class}"
                not in self.parsing_register[RegisterUnit.Type].relations
            ), self.error_msg(
                f'Relation "{match.group()}" is duplicated under the type {type_name[type_name.index(".") + 1:]}'
                if self.parsing_register[RegisterUnit.Type].spg_type_enum
                != SpgTypeEnum.Concept
                else f'Relation "{match.group()}" is already defined by keyword autoRelate'
                f'under the {type_name[type_name.index(".") + 1:]}'
            )

            predicate = Relation(name=predicate_name, object_type_name=predicate_class)
            if predicate_class in self.types:
                predicate.object_spg_type = self.types[predicate_class].spg_type_enum
            self.parsing_register[RegisterUnit.Type].add_relation(predicate)
            self.save_register(RegisterUnit.Relation, predicate)
        predicate.name_zh = predicate_name_zh

    def parse_property_meta(self, expression):
        """
        parse the property meta definition of SPG type
        """

        match = re.match(r"^(desc|properties|constraint|rule):\s*?(.*)$", expression)
        assert match, self.error_msg(
            "Unrecognized expression, expect desc:|properties:|constraint:|rule:"
        )

        property_meta = match.group(1)
        meta_value = match.group(2)

        if property_meta == "desc" and len(meta_value) > 0:
            if self.parsing_register[RegisterUnit.SubProperty] is not None:
                self.parsing_register[
                    RegisterUnit.SubProperty
                ].desc = meta_value.strip()
            elif self.parsing_register[RegisterUnit.Property] is not None:
                self.parsing_register[RegisterUnit.Property].desc = meta_value.strip()

        elif property_meta == "constraint":
            if self.parsing_register[RegisterUnit.SubProperty] is not None:
                self.parse_constraint_for_property(
                    meta_value, self.parsing_register[RegisterUnit.SubProperty]
                )
            elif self.parsing_register[RegisterUnit.Property] is not None:
                self.parse_constraint_for_property(
                    meta_value, self.parsing_register[RegisterUnit.Property]
                )

        elif property_meta == "properties":
            self.save_register(
                RegisterUnit.SubProperty, Property(name="_", object_type_name="Thing")
            )

        elif property_meta == "rule":
            self.parse_predicate_rule(meta_value.lstrip(), RegisterUnit.Property)

    def parse_relation_meta(self, expression):
        """
        parse the relation meta definition of SPG type
        """

        match = re.match(r"^(desc|properties|rule):\s*?(.*)$", expression)
        assert match, self.error_msg(
            "Unrecognized expression, expect desc:|properties:|rule:"
        )

        property_meta = match.group(1)
        meta_value = match.group(2)

        if property_meta == "desc" and len(meta_value) > 0:
            self.parsing_register[RegisterUnit.Relation].desc = meta_value.strip()

        elif property_meta == "properties":
            self.save_register(
                RegisterUnit.SubProperty, Property(name="_", object_type_name="Thing")
            )

        elif property_meta == "rule":
            self.parse_predicate_rule(meta_value.lstrip(), RegisterUnit.Relation)

    def parsing_dispatch(self, expression, parsing_level):
        if parsing_level == IndentLevel.Type.value:
            self.parse_type(expression)

        elif parsing_level == IndentLevel.TypeMeta.value:
            self.parse_type_meta(expression)

        elif parsing_level == IndentLevel.Predicate.value:
            self.parse_predicate(expression)

        elif parsing_level == IndentLevel.PredicateMeta.value:
            if self.parsing_register[RegisterUnit.Property] is not None:
                self.parse_property_meta(expression)

            else:
                self.parse_relation_meta(expression)

        elif parsing_level == IndentLevel.SubProperty.value:
            self.parse_predicate(expression)

        elif parsing_level == IndentLevel.SubPropertyMeta.value:
            self.parse_property_meta(expression)

    def parse_predicate_rule(self, rule, key):
        """
        parse the logic rule for property/relation
        """

        strip_rule = rule
        if strip_rule.startswith("[["):
            self.rule_quote_predicate = self.parsing_register[key]
            self.rule_quote_open = True
            if len(strip_rule) > 2:
                self.rule_quote_predicate.logical_rule = strip_rule[2].lstrip()
            else:
                self.rule_quote_predicate.logical_rule = ""
        else:
            self.parsing_register[key].logical_rule = rule

    def parse_constraint_for_property(self, expression, prop):
        """
        parse the constraint definition of property
        """

        if len(expression) == 0:
            return

        pattern = re.compile(r"(Enum|Regular)\s*?=\s*?\"([^\"]+)\"", re.IGNORECASE)
        matches = re.findall(pattern, expression)
        if matches:
            for group in matches:
                if group[0].lower() == "enum":
                    enum_values = group[1].split(",")
                    strip_enum_values = list()
                    for ev in enum_values:
                        strip_enum_values.append(ev.strip())
                    prop.add_constraint(ConstraintTypeEnum.Enum, strip_enum_values)

                elif group[0].lower() == "regular":
                    prop.add_constraint(ConstraintTypeEnum.Regular, group[1])

        expression = re.sub(r"(Enum|Regular)\s*?=\s*?\"([^\"]+)\"", "", expression)
        array = expression.split(",")
        for cons in array:
            cons = cons.strip()
            if cons.lower() == "multivalue":
                prop.add_constraint(ConstraintTypeEnum.MultiValue)

            elif cons.lower() == "notnull":
                prop.add_constraint(ConstraintTypeEnum.NotNull)

    def complete_rule(self, rule):
        """
        Auto generate define statement and append namespace to the entity name
        """

        pattern = re.compile(r"Define\s*\(", re.IGNORECASE)
        match = pattern.match(rule.strip())
        if not match:
            subject_name = self.parsing_register[RegisterUnit.Type].name
            predicate = None
            if self.parsing_register[RegisterUnit.Property] is not None:
                predicate = self.parsing_register[RegisterUnit.Property]
            elif self.parsing_register[RegisterUnit.Relation] is not None:
                predicate = self.parsing_register[RegisterUnit.Relation]
            head = (
                f"Define (s:{subject_name})-[p:{predicate.name}]->(o:{predicate.object_type_name})"
                + " {\n"
            )
            rule = head + rule
            rule += "\n}"

        pattern = re.compile(r"\(([\w\s]*?:)(`?[\w\s\.]+)`?/?[^)]*?\)", re.IGNORECASE)
        matches = re.findall(pattern, rule)
        replace_list = []
        if matches:
            for group in matches:
                if "." in group[1] or group[1].lower() in ["integer", "text", "float"]:
                    continue
                replace_list.append(
                    (
                        f"({group[0]}{group[1]}",
                        f"({group[0]}{self.namespace}.{group[1].strip()}"
                        if "`" not in group[1]
                        else f"({group[0]}`{self.namespace}.{group[1].replace('`', '').strip()}",
                    )
                )
        if len(replace_list) > 0:
            for t in replace_list:
                rule = rule.replace(t[0], t[1])

        return rule.strip()

    def preload_types(self, lines: list):
        """
        Pre analyze the script to obtain defined types
        """

        for line in lines:
            type_match = re.match(
                r"^([a-zA-Z0-9\.]+)\((\w+)\):\s*?([a-zA-Z0-9,]+)$", line
            )
            if type_match:
                self.defined_types[type_match.group(1)] = type_match.group(3).strip()
                continue
            sub_type_match = re.match(
                r"^([a-zA-Z0-9]+)\((\w+)\)\s*?->\s*?([a-zA-Z0-9\.]+):$", line
            )
            if sub_type_match:
                self.defined_types[sub_type_match.group(1)] = sub_type_match.group(
                    3
                ).strip()

    def load_script(self):
        """
        Load and then parse the script file
        """

        file = open(self.schema_file, "r", encoding="utf-8")
        lines = file.read().splitlines()
        self.preload_types(lines)
        for line in lines:
            self.current_line_num += 1
            strip_line = line.strip()
            # replace tabs with two spaces
            line = line.replace("\t", "  ")
            if strip_line == "" or strip_line.startswith("#"):
                # skip empty or comments line
                continue

            if self.rule_quote_open:
                # process the multi-line assignment [[ .... ]]
                right_strip_line = line.rstrip()
                if strip_line.endswith("]]"):
                    self.rule_quote_open = False
                    if len(right_strip_line) > 2:
                        self.rule_quote_predicate.logical_rule += right_strip_line[
                            : len(right_strip_line) - 2
                        ]
                    self.rule_quote_predicate.logical_rule = self.complete_rule(
                        self.rule_quote_predicate.logical_rule
                    )

                else:
                    self.rule_quote_predicate.logical_rule += line + "\n"
                continue

            indent_count = len(line) - len(line.lstrip())
            if indent_count == 0:
                # the line without indent is namespace definition or a type definition
                self.adjust_parsing_level(0)

            elif indent_count > self.last_indent_level:
                # the line is the sub definition of the previous line
                self.adjust_parsing_level(1)

            elif indent_count < self.last_indent_level:
                # finish current indent parsing
                backward_step = None
                for i in range(0, len(self.indent_level_pos)):
                    if indent_count == self.indent_level_pos[i]:
                        backward_step = i - self.current_parsing_level
                        break
                assert backward_step, self.error_msg(
                    f"Invalid indentation, please align with the previous definition"
                )

                if backward_step != 0:
                    self.adjust_parsing_level(backward_step)

            self.parsing_dispatch(strip_line, self.current_parsing_level)
            self.last_indent_level = indent_count
            self.indent_level_pos[self.current_parsing_level] = indent_count

    def is_internal_property(self, prop: Property, spg_type: SpgTypeEnum):
        if spg_type == SpgTypeEnum.Entity or spg_type == SpgTypeEnum.Standard:
            return prop in self.entity_internal_property

        elif spg_type == SpgTypeEnum.Concept:
            return prop in self.concept_internal_property

        elif spg_type == SpgTypeEnum.Event:
            return prop in self.event_internal_property

    def sync_schema(self):
        return self.diff_and_sync(False)

    def print_diff(self):
        self.diff_and_sync(True)

    def diff_sub_property(self, new, old, old_type_name, old_property, new_property):
        need_update = False
        inherited_type = self.get_inherited_type(old_type_name)
        for prop in old:
            if not old_property.inherited and prop not in new:
                assert inherited_type is None, self.error_msg(
                    f'"{old_type_name} was inherited by other type, such as "{inherited_type}". Prohibit property alteration!'
                )

                old[prop].alter_operation = AlterOperationEnum.Delete
                need_update = True
                print(
                    f"Delete sub property: [{old_type_name}] {old_property.name}.{prop}"
                )

        for prop, o in new.items():
            if prop not in old and not new_property.inherited:
                assert inherited_type is None, self.error_msg(
                    f'"{old_type_name} was inherited by other type, such as "{inherited_type}". Prohibit property alteration!'
                )

                old_property.add_sub_property(new[prop])
                need_update = True
                print(
                    f"Create sub property: [{old_type_name}] {old_property.name}.{prop}"
                )

            elif old[prop].object_type_name != new[prop].object_type_name:
                assert inherited_type is None, self.error_msg(
                    f'"{old_type_name} was inherited by other type, such as "{inherited_type}". Prohibit property alteration!'
                )
                assert not old_property.inherited, self.error_msg(
                    f"{old_type_name}] {old_property.name}.{prop} is inherited sub property, deny modify"
                )

                old[prop].alter_operation = AlterOperationEnum.Delete
                old_property.add_sub_property(new[prop])
                need_update = True
                print(
                    f"Recreate sub property: [{old_type_name}] {old_property.name}.{prop}"
                )

            elif old[prop] != new[prop]:
                assert inherited_type is None, self.error_msg(
                    f'"{old_type_name} was inherited by other type, such as "{inherited_type}". Prohibit property alteration!'
                )
                assert not old_property.inherited, self.error_msg(
                    f"{old_type_name}] {old_property.name}.{prop} is inherited property, deny modify"
                )

                old[prop].overwritten_by(o)
                old[prop].alter_operation = AlterOperationEnum.Update
                need_update = True
                print(f"Update property: [{old_type_name}] {old_property.name}.{prop}")
        return need_update

    def get_inherited_type(self, type_name):
        for spg_type in self.types:
            if self.types[spg_type].parent_type_name == type_name:
                return spg_type
        return None

    def diff_and_sync(self, print_only):
        """
        Get the schema diff and then sync to graph storage
        """
        schema = SchemaClient()
        session = schema.create_session()

        # generate the delete list of spg type
        for spg_type in session.spg_types:
            if not spg_type.startswith("STD.") and not spg_type.startswith(
                f"{self.namespace}."
            ):
                continue
            unique_id = session.spg_types[spg_type]._rest_model.ontology_id.unique_id
            if spg_type in self.internal_type and unique_id < 1000:
                continue

            if spg_type not in self.types:
                session.delete_type(session.get(spg_type))
                print(f"Delete type: {spg_type}")

        for spg_type in self.types:
            # generate the creation list of spg type
            if not spg_type.startswith("STD.") and not spg_type.startswith(
                f"{self.namespace}."
            ):
                continue
            if spg_type not in session.spg_types:
                session.create_type(self.types[spg_type])
                print(f"Create type: {spg_type}")
                relations = self.types[spg_type].relations
                if len(relations) > 0:
                    for rel in relations:
                        print(f'Create relation: [{spg_type}] {rel.split("_")[0]}')

            else:
                # generate the update list
                new_type = self.types[spg_type]
                old_type = session.get(spg_type)

                assert (
                    new_type.spg_type_enum == old_type.spg_type_enum
                    and new_type.parent_type_name == old_type.parent_type_name
                ), self.error_msg(
                    f"Cannot alter the type definition or its parent type of {new_type.name}. "
                    "if you still want to make change, please delete it first then re-create it."
                )

                need_update = False
                if new_type.desc != old_type.desc:
                    old_type.desc = new_type.desc
                    need_update = True

                if new_type.name_zh != old_type.name_zh:
                    old_type.name_zh = new_type.name_zh
                    need_update = True

                if new_type.spg_type_enum == SpgTypeEnum.Concept:
                    assert (
                        new_type.hypernym_predicate == old_type.hypernym_predicate
                    ), self.error_msg(
                        f"Cannot alter the hypernym predicate of {new_type.name}. "
                        "if you still want to make change, please delete it first then re-create it."
                    )

                if new_type.spg_type_enum == SpgTypeEnum.Standard:
                    assert old_type.spreadable == new_type.spreadable, self.error_msg(
                        f"Cannot alter the spreadable value of {new_type.name}. "
                        f"if you still want to make change, "
                        "please delete the definition first and then re-create it."
                    )

                    if old_type.constraint != new_type.constraint:
                        old_type.constraint = new_type.constraint
                        need_update = True
                        print(f"Update standard type constraint: {spg_type}")

                inherited_type = self.get_inherited_type(new_type.name)
                for prop in old_type.properties:
                    if (
                        not old_type.properties[prop].inherited
                        and prop not in new_type.properties
                        and not self.is_internal_property(prop, new_type.spg_type_enum)
                    ):
                        assert (
                            prop != "subject"
                            and old_type.properties[prop].property_group
                            != PropertyGroupEnum.Subject
                        ), self.error_msg(
                            "The subject property of event type cannot be deleted"
                        )
                        assert inherited_type is None, self.error_msg(
                            f'"{new_type.name} was inherited by other type, such as "{inherited_type}". Prohibit property alteration!'
                        )

                        old_type.properties[
                            prop
                        ].alter_operation = AlterOperationEnum.Delete
                        need_update = True
                        print(f"Delete property: [{new_type.name}] {prop}")

                for prop, o in new_type.properties.items():

                    if (
                        prop not in old_type.properties
                        and not self.is_internal_property(prop, new_type.spg_type_enum)
                        and not o.inherited
                    ):
                        assert inherited_type is None, self.error_msg(
                            f'"{new_type.name} was inherited by other type, such as "{inherited_type}". Prohibit property alteration!'
                        )

                        old_type.add_property(new_type.properties[prop])
                        need_update = True
                        print(f"Create property: [{new_type.name}] {prop}")

                    elif (
                        old_type.properties[prop].object_type_name
                        != new_type.properties[prop].object_type_name
                    ):
                        assert inherited_type is None, self.error_msg(
                            f'"{new_type.name} was inherited by other type, such as "{inherited_type}". Prohibit property alteration!'
                        )
                        assert not old_type.properties[prop].inherited, self.error_msg(
                            f"{new_type.name}] {prop} is inherited property, deny modify"
                        )

                        old_type.properties[
                            prop
                        ].alter_operation = AlterOperationEnum.Delete
                        old_type.add_property(new_type.properties[prop])
                        need_update = True
                        print(f"Recreate property: [{new_type.name}] {prop}")

                    elif (
                        old_type.properties[prop].sub_properties
                        != new_type.properties[prop].sub_properties
                    ):
                        need_update = self.diff_sub_property(
                            new_type.properties[prop].sub_properties,
                            old_type.properties[prop].sub_properties,
                            old_type.name,
                            old_type.properties[prop],
                            new_type.properties[prop],
                        )
                        if need_update:
                            old_type.properties[
                                prop
                            ].alter_operation = AlterOperationEnum.Update

                    elif old_type.properties[prop] != new_type.properties[prop]:
                        assert inherited_type is None, self.error_msg(
                            f'"{new_type.name} was inherited by other type, such as "{inherited_type}". Prohibit property alteration!'
                        )
                        assert not old_type.properties[prop].inherited, self.error_msg(
                            f"{new_type.name}] {prop} is inherited property, deny modify"
                        )

                        old_type.properties[prop].overwritten_by(o)
                        old_type.properties[
                            prop
                        ].alter_operation = AlterOperationEnum.Update
                        need_update = True
                        print(f"Update property: [{new_type.name}] {prop}")

                for relation in new_type.relations:
                    p_name = relation.split("_")[0]
                    if (
                        relation not in old_type.relations
                        or old_type.relations[relation].object_type_name
                        != new_type.relations[relation].object_type_name
                    ):
                        assert inherited_type is None, self.error_msg(
                            f'"{new_type.name} was inherited by other type, such as "{inherited_type}". Prohibit relation alteration!'
                        )
                        old_type.add_relation(new_type.relations[relation])
                        need_update = True
                        print(f"Create relation: [{new_type.name}] {p_name}")

                    elif (
                        old_type.relations[relation].sub_properties
                        != new_type.relations[relation].sub_properties
                    ):
                        need_update = self.diff_sub_property(
                            new_type.relations[relation].sub_properties,
                            old_type.relations[relation].sub_properties,
                            old_type.name,
                            old_type.relations[relation],
                            new_type.relations[relation],
                        )
                        if need_update:
                            assert inherited_type is None, self.error_msg(
                                f'"{new_type.name} was inherited by other type, such as "{inherited_type}". Prohibit relation alteration!'
                            )
                            old_type.relations[
                                relation
                            ].alter_operation = AlterOperationEnum.Update

                    elif old_type.relations[relation] != new_type.relations[relation]:
                        assert inherited_type is None, self.error_msg(
                            f'"{new_type.name} was inherited by other type, such as "{inherited_type}". Prohibit relation alteration!'
                        )
                        assert not old_type.relations[
                            relation
                        ].inherited, self.error_msg(
                            f"{new_type.name}] {p_name} is inherited relation, deny modify"
                        )

                        old_type.relations[relation].overwritten_by(
                            new_type.relations[relation]
                        )
                        old_type.relations[
                            relation
                        ].alter_operation = AlterOperationEnum.Update
                        need_update = True
                        print(f"Update relation: [{new_type.name}] {relation}")

                for relation, o in old_type.relations.items():
                    p_name = relation.split("_")[0]
                    if o.inherited or p_name in new_type.properties or o.is_dynamic:
                        # skip the inherited and semantic relation
                        continue
                    if (
                        relation not in new_type.relations
                        and not o.inherited
                        and not o.is_dynamic
                        and not (
                            new_type.spg_type_enum == SpgTypeEnum.Concept
                            and p_name
                            in [member.value for member in HypernymPredicateEnum]
                        )
                    ):
                        assert inherited_type is None, self.error_msg(
                            f'"{new_type.name} was inherited by other type, such as "{inherited_type}". Prohibit relation alteration!'
                        )
                        old_type.relations[
                            relation
                        ].alter_operation = AlterOperationEnum.Delete
                        need_update = True
                        print(f"Delete relation: [{new_type.name}] {p_name}")

                if need_update:
                    session.update_type(old_type)
        if not print_only:
            session.commit()
        if session._alter_spg_types:
            return True
        return False

    def export_schema_python(self, filename):
        """
        Export the schema helper class in python
        You can import the exported class in your code to obtain the code prompt in IDE
        """

        schema = SchemaClient()
        session = schema.create_session()
        assert len(self.namespace) > 0, "Schema is invalid"

        spg_types = []
        for spg_type_name in sorted(session.spg_types):
            if (
                spg_type_name.startswith("STD.")
                or not spg_type_name.startswith(self.namespace)
                or spg_type_name in self.internal_type
            ):
                continue

            sub_properties = {}
            properties = set()
            for prop, prop_type in session.get(spg_type_name).properties.items():
                if len(prop_type.sub_properties) > 0:
                    sub_properties[prop] = set()
                    for sub_prop in prop_type.sub_properties:
                        sub_properties[prop].add(sub_prop)
                else:
                    properties.add(prop)

            relations = set()
            relation_sub_properties = {}
            hyp_predicate = [member.value for member in HypernymPredicateEnum]
            for relation, relation_type in session.get(spg_type_name).relations.items():
                rel = relation.split("_")[0]
                if (
                    rel in relations
                    or rel in hyp_predicate
                    or rel in session.get(spg_type_name).properties
                ):
                    continue

                if len(relation_type.sub_properties) > 0:
                    relation_sub_properties[rel] = set()
                    for sub_prop in relation_type.sub_properties:
                        relation_sub_properties[rel].add(sub_prop)
                else:
                    relations.add(rel)

            spg_types.append(
                {
                    "name": spg_type_name.split(".")[1],
                    "properties": properties,
                    "sub_properties": sub_properties,
                    "relations": relations,
                    "relation_sub_properties": relation_sub_properties,
                }
            )

        metadata = {"namespace": self.namespace, "spg_types": spg_types}

        from knext.common.template import render_template

        render_template(Path(filename).parent, Path(filename).name, **metadata)



# coding: utf-8
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


"""
    knext

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from knext.common.rest.configuration import Configuration


class LogicalRule(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        "code": "RuleCode",
        "name": "str",
        "version": "int",
        "is_master": "bool",
        "atatus": "str",
        "content": "str",
        "creator": "UserInfo",
    }

    attribute_map = {
        "code": "code",
        "name": "name",
        "version": "version",
        "is_master": "isMaster",
        "atatus": "atatus",
        "content": "content",
        "creator": "creator",
    }

    def __init__(
        self,
        code=None,
        name=None,
        version=None,
        is_master=None,
        atatus=None,
        content=None,
        creator=None,
        local_vars_configuration=None,
    ):  # noqa: E501
        """LogicalRule - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._code = None
        self._name = None
        self._version = None
        self._is_master = None
        self._atatus = None
        self._content = None
        self._creator = None
        self.discriminator = None

        if code is not None:
            self.code = code
        if name is not None:
            self.name = name
        if version is not None:
            self.version = version
        if is_master is not None:
            self.is_master = is_master
        if atatus is not None:
            self.atatus = atatus
        if content is not None:
            self.content = content
        if creator is not None:
            self.creator = creator

    @property
    def code(self):
        """Gets the code of this LogicalRule.  # noqa: E501


        :return: The code of this LogicalRule.  # noqa: E501
        :rtype: RuleCode
        """
        return self._code

    @code.setter
    def code(self, code):
        """Sets the code of this LogicalRule.


        :param code: The code of this LogicalRule.  # noqa: E501
        :type: RuleCode
        """

        self._code = code

    @property
    def name(self):
        """Gets the name of this LogicalRule.  # noqa: E501


        :return: The name of this LogicalRule.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this LogicalRule.


        :param name: The name of this LogicalRule.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def version(self):
        """Gets the version of this LogicalRule.  # noqa: E501


        :return: The version of this LogicalRule.  # noqa: E501
        :rtype: int
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this LogicalRule.


        :param version: The version of this LogicalRule.  # noqa: E501
        :type: int
        """

        self._version = version

    @property
    def is_master(self):
        """Gets the is_master of this LogicalRule.  # noqa: E501


        :return: The is_master of this LogicalRule.  # noqa: E501
        :rtype: bool
        """
        return self._is_master

    @is_master.setter
    def is_master(self, is_master):
        """Sets the is_master of this LogicalRule.


        :param is_master: The is_master of this LogicalRule.  # noqa: E501
        :type: bool
        """

        self._is_master = is_master

    @property
    def atatus(self):
        """Gets the atatus of this LogicalRule.  # noqa: E501


        :return: The atatus of this LogicalRule.  # noqa: E501
        :rtype: str
        """
        return self._atatus

    @atatus.setter
    def atatus(self, atatus):
        """Sets the atatus of this LogicalRule.


        :param atatus: The atatus of this LogicalRule.  # noqa: E501
        :type: str
        """
        allowed_values = ["INIT", "GRAY", "PROD", "OFF", "DEL"]  # noqa: E501
        if (
            self.local_vars_configuration.client_side_validation
            and atatus not in allowed_values
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `atatus` ({0}), must be one of {1}".format(  # noqa: E501
                    atatus, allowed_values
                )
            )

        self._atatus = atatus

    @property
    def content(self):
        """Gets the content of this LogicalRule.  # noqa: E501


        :return: The content of this LogicalRule.  # noqa: E501
        :rtype: str
        """
        return self._content

    @content.setter
    def content(self, content):
        """Sets the content of this LogicalRule.


        :param content: The content of this LogicalRule.  # noqa: E501
        :type: str
        """

        self._content = content

    @property
    def creator(self):
        """Gets the creator of this LogicalRule.  # noqa: E501


        :return: The creator of this LogicalRule.  # noqa: E501
        :rtype: UserInfo
        """
        return self._creator

    @creator.setter
    def creator(self, creator):
        """Sets the creator of this LogicalRule.


        :param creator: The creator of this LogicalRule.  # noqa: E501
        :type: UserInfo
        """

        self._creator = creator

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, LogicalRule):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, LogicalRule):
            return True

        return self.to_dict() != other.to_dict()



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from abc import ABC
from enum import Enum
from functools import cmp_to_key
from typing import Union

from knext.common.base.component import Component


class ComponentTypeEnum(str, Enum):
    CSVReader = "CSV_SOURCE"
    LLMBasedExtractor = "LLM_BASED_EXTRACT"
    UserDefinedExtractor = "USER_DEFINED_EXTRACT"
    SPGTypeMapping = "SPG_TYPE_MAPPING"
    RelationMapping = "RELATION_MAPPING"
    SubGraphMapping = "SUBGRAPH_MAPPING"
    KGWriter = "GRAPH_SINK"


class BuilderComponent(Component, ABC):
    """
    Abstract base class for all builder component.
    """

    @property
    def type(self):
        return ComponentTypeEnum.__members__[self.__class__.__name__].value


class SourceReader(BuilderComponent, ABC):
    """
    Abstract base class for all source reader component.
    """

    @property
    def upstream_types(self):
        return None

    @property
    def downstream_types(self):
        return Union[SPGExtractor, Mapping]


class SPGExtractor(BuilderComponent, ABC):
    """
    Abstract base class for all SPG extractor component.
    """

    @property
    def upstream_types(self):
        return Union[SourceReader, SPGExtractor]

    @property
    def downstream_types(self):
        return Union[SPGExtractor, Mapping]


class Mapping(BuilderComponent, ABC):
    """
    Abstract base class for all mapping component.
    """

    @property
    def upstream_types(self):
        return Union[SourceReader, SPGExtractor]

    @property
    def downstream_types(self):
        return Union[SinkWriter]

    @staticmethod
    def sort_by_dependency(mappings: list):

        from knext.builder.component import SPGTypeMapping

        def comparator(x: SPGTypeMapping, y: SPGTypeMapping):
            if x.spg_type_name in y.dependencies:
                return -1
            elif y.spg_type_name in x.dependencies:
                return 1
            else:
                return 0

        from knext.builder import rest

        if len(mappings) == 1:
            return rest.SpgTypeMappingNodeConfigs(
                mapping_node_configs=[m.to_rest().node_config for m in mappings]
            )
        mappings = sorted(mappings, key=cmp_to_key(comparator))
        return rest.SpgTypeMappingNodeConfigs(
            mapping_node_configs=[m.to_rest().node_config for m in mappings]
        )


class SinkWriter(BuilderComponent, ABC):
    """
    Abstract base class for all sink writer component.
    """

    @property
    def upstream_types(self):
        return Union[Mapping]

    @property
    def downstream_types(self):
        return None



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import pprint
import typing
from abc import ABC
from enum import Enum
from typing import Type, Union, List, Dict, Optional

from knext.schema import rest

ROOT_TYPE_UNIQUE_NAME = "Thing"


class SpgTypeEnum(str, Enum):
    Basic = "BASIC_TYPE"
    Standard = "STANDARD_TYPE"
    Entity = "ENTITY_TYPE"
    Event = "EVENT_TYPE"
    Concept = "CONCEPT_TYPE"


class BasicTypeEnum(str, Enum):
    Text = "Text"
    Integer = "Integer"
    Float = "Float"


class PropertyGroupEnum(str, Enum):
    Time = "TIME"
    Subject = "SUBJECT"
    Object = "OBJECT"
    Loc = "LOC"


class ConstraintTypeEnum(str, Enum):
    NotNull = "NOT_NULL"
    MultiValue = "MULTI_VALUE"
    Enum = "ENUM"
    Regular = "REGULAR"


class HypernymPredicateEnum(str, Enum):
    IsA = "isA"
    LocateAt = "locateAt"
    MannerOf = "mannerOf"


class AlterOperationEnum(str, Enum):
    Create = "CREATE"
    Update = "UPDATE"
    Delete = "DELETE"


def iter_init(klass):
    """Initialize a REST model."""
    instance = klass()
    for attr, attr_type in klass.openapi_types.items():
        if hasattr(rest, attr_type):
            attr_klass = getattr(rest, attr_type)
            attr_instance = iter_init(attr_klass)
            setattr(instance, attr, attr_instance)
        elif attr_type.startswith("list["):
            setattr(instance, attr, [])
        else:
            pass

    return instance


class BaseProperty(ABC):
    """Base class of `Property` and `Relation`."""

    _rest_model: Union[rest.Relation, rest.Property]

    def __init__(
        self,
        name=None,
        object_type_name=None,
        name_zh=None,
        desc=None,
        property_group=None,
        sub_properties=None,
        constraint=None,
        logical_rule=None,
        **kwargs,
    ):
        if "rest_model" in kwargs:
            self._rest_model = kwargs["rest_model"]
        else:
            self._init_rest_model(
                name=name,
                object_type_name=object_type_name,
                name_zh=name_zh,
                desc=desc,
                property_group=property_group,
                sub_properties=sub_properties,
                constraint=constraint,
                logical_rule=logical_rule,
            )

    def _init_rest_model(self, **kwargs):
        """Init a BaseProperty object."""
        super_klass = self.__class__.__name__
        self._rest_model = iter_init(getattr(rest, super_klass))
        for param, value in kwargs.items():
            setattr(self, param, value)

    @property
    def name(self) -> str:
        """Gets the name of this Property/Relation.  # noqa: E501


        :return: The name of this Property/Relation.  # noqa: E501
        :rtype: str
        """
        return self._rest_model.basic_info.name.name

    @name.setter
    def name(self, name: str):
        """Sets the name of this Property/Relation.


        :param name: The name of this Property/Relation.  # noqa: E501
        :type: str
        """

        self._rest_model.basic_info.name.name = name

    @property
    def object_type_name(self) -> str:
        """Gets the object_type_name of this Property/Relation.  # noqa: E501


        :return: The object_type_name of this Property/Relation.  # noqa: E501
        :rtype: str
        """
        return self._rest_model.object_type_ref.basic_info.name.name

    @object_type_name.setter
    def object_type_name(self, object_type_name: str):
        """Sets the object_type_name of this Property/Relation.


        :param object_type_name: The object_type_name of this Property/Relation.  # noqa: E501
        :type: str
        """

        self._rest_model.object_type_ref.basic_info.name.name = object_type_name

    @property
    def object_type_name_zh(self) -> str:
        """Gets the object_type_name_zh of this Property/Relation.  # noqa: E501


        :return: The object_type_name_zh of this Property/Relation.  # noqa: E501
        :rtype: str
        """
        return self._rest_model.object_type_ref.basic_info.name_zh

    @object_type_name_zh.setter
    def object_type_name_zh(self, object_type_name_zh: str):
        """Sets the object_type_name_zh of this Property/Relation.


        :param object_type_name_zh: The object_type_name_zh of this Property/Relation.  # noqa: E501
        :type: str
        """
        self._rest_model.object_type_ref.basic_info.name_zh = object_type_name_zh

    @property
    def inherited(self) -> bool:
        """Gets the `inherited` of this Property/Relation.  # noqa: E501


        :return: The `inherited` of this Property/Relation.  # noqa: E501
        :rtype: bool
        """
        return self._rest_model.inherited

    @inherited.setter
    def inherited(self, inherited: bool):
        """Sets the `inherited` of this Property/Relation.


        :param inherited: The `inherited` of this Property/Relation.  # noqa: E501
        :type: bool
        """

        if inherited is None:
            return

        self._rest_model.inherited = inherited

    @property
    def object_spg_type(self) -> Optional[SpgTypeEnum]:
        """Gets the object_spg_type of this Property/Relation.  # noqa: E501


        :return: The object_spg_type of this Property/Relation.  # noqa: E501
        :rtype: str
        """
        spg_type_enum = self._rest_model.object_type_ref.spg_type_enum
        return SpgTypeEnum(spg_type_enum) if spg_type_enum else None

    @object_spg_type.setter
    def object_spg_type(self, object_spg_type: SpgTypeEnum):
        """Sets the object_spg_type of this Property/Relation.


        :param object_spg_type: The object_spg_type of this Property/Relation.  # noqa: E501
        :type: str
        """

        if object_spg_type is None:
            return

        self._rest_model.object_type_ref.spg_type_enum = object_spg_type

    @property
    def name_zh(self) -> str:
        """Gets the name_zh of this Property/Relation.  # noqa: E501


        :return: The name_zh of this Property/Relation.  # noqa: E501
        :rtype: str
        """
        return self._rest_model.basic_info.name_zh

    @name_zh.setter
    def name_zh(self, name_zh: str):
        """Sets the name_zh of this Property/Relation.


        :param name_zh: The name_zh of this Property/Relation.  # noqa: E501
        :type: str
        """
        if name_zh is None:
            return

        self._rest_model.basic_info.name_zh = name_zh

    @property
    def desc(self) -> str:
        """Gets the desc of this Property/Relation.  # noqa: E501


        :return: The desc of this Property/Relation.  # noqa: E501
        :rtype: str
        """
        return self._rest_model.basic_info.desc

    @desc.setter
    def desc(self, desc: str):
        """Sets the desc of this Property/Relation.


        :param desc: The desc of this Property/Relation.  # noqa: E501
        :type: str
        """
        if desc is None:
            return

        self._rest_model.basic_info.desc = desc

    @property
    def property_group(self) -> Optional[PropertyGroupEnum]:
        """Gets the property_group of this Property/Relation.  # noqa: E501


        :return: The property_group of this Property/Relation.  # noqa: E501
        :rtype: str
        """
        property_group = self._rest_model.advanced_config.property_group
        return PropertyGroupEnum(property_group) if property_group else None

    @property_group.setter
    def property_group(self, property_group: PropertyGroupEnum):
        """Sets the property_group of this Property/Relation.


        :param property_group: The property_group of this Property/Relation.  # noqa: E501
        :type: str
        """
        if property_group is None:
            return

        self._rest_model.advanced_config.property_group = property_group

    @property
    def sub_properties(self) -> Dict[str, Type["Property"]]:
        """Gets the sub_properties of this Property/Relation.  # noqa: E501


        :return: The sub_properties of this Property/Relation.  # noqa: E501
        :rtype: dict
        """
        if self._rest_model.advanced_config.sub_properties is None:
            return {}
        from knext.schema.model.property import Property

        sub_properties = {}
        for sub_property in self._rest_model.advanced_config.sub_properties:
            sub_properties[sub_property.basic_info.name.name] = Property(
                name=sub_property.basic_info.name.name,
                object_type_name=sub_property.object_type_ref.basic_info.name.name,
                rest_model=sub_property,
            )
        return sub_properties

    @sub_properties.setter
    def sub_properties(self, sub_properties: List["Property"]):
        """Sets the sub_properties of this Property/Relation.


        :param sub_properties: The sub_properties of this Property/Relation.  # noqa: E501
        :type: list[Property]
        """

        if sub_properties is None:
            return

        self._rest_model.advanced_config.sub_properties = [
            prop.to_rest() for prop in sub_properties
        ]

    def add_sub_property(self, sub_property: Type["Property"]):
        """Adds a sub_property to this Property/Relation.


        :param sub_property: The sub_property to add.
        :type sub_property: Property
        """

        if self._rest_model.advanced_config.sub_properties is None:
            self._rest_model.advanced_config.sub_properties = None
        sub_property.alter_operation = AlterOperationEnum.Create
        self._rest_model.advanced_config.sub_properties.append(sub_property.to_rest())
        return self

    @property
    def constraint(self) -> Dict[ConstraintTypeEnum, Union[str, list]]:
        """Gets the constraint of this Property.  # noqa: E501


        :return: The constraint of this Property.  # noqa: E501
        :rtype: dict
        """
        if self._rest_model.advanced_config.constraint is None:
            return {}
        constraint = {}
        for item in self._rest_model.advanced_config.constraint.constraint_items:
            if item.constraint_type_enum == ConstraintTypeEnum.Enum:
                value = item.enum_values
            elif item.constraint_type_enum == ConstraintTypeEnum.Regular:
                value = item.regular_pattern
            else:
                value = None
            constraint[item.constraint_type_enum] = value
        return constraint

    @constraint.setter
    def constraint(self, constraint: Dict[ConstraintTypeEnum, Union[str, list]]):
        """Sets the constraint of this Property.


        :param constraint: The constraint of this Property.  # noqa: E501
        :type: dict
        """
        if constraint is None:
            return
        self._rest_model.advanced_config.constraint = rest.Constraint(
            constraint_items=[]
        )
        for type, value in constraint.items():
            self.add_constraint(type, value)

    def add_constraint(self, type: ConstraintTypeEnum, value: Union[str, list] = None):
        """Adds a constraint to this Property.


        :param type: The type of constraint to add.
        :type type: ConstraintTypeEnum
        :param value: The value(s) of the constraint. Optional.
        :type value: str or list, optional
        """

        if self._rest_model.advanced_config.constraint is None:
            self._rest_model.advanced_config.constraint = rest.Constraint(
                constraint_items=[]
            )
        if type == ConstraintTypeEnum.Enum:
            if not isinstance(value, list):
                raise ValueError("Invalid enum format.")
            constraint_item = rest.EnumConstraint(enum_values=value)
        elif type == ConstraintTypeEnum.Regular:
            constraint_item = rest.RegularConstraint(regular_pattern=value)
        else:
            constraint_item = rest.BaseConstraintItem(type)
        self._rest_model.advanced_config.constraint.constraint_items.append(
            constraint_item
        )
        return self

    @property
    def logical_rule(self) -> str:
        """Gets the logical_rule of this Property/Relation.  # noqa: E501


        :return: The logical_rule of this Property/Relation.  # noqa: E501
        :rtype: str
        """
        if self._rest_model.advanced_config.logical_rule is None:
            return ""
        return self._rest_model.advanced_config.logical_rule.content

    @logical_rule.setter
    def logical_rule(self, logical_rule: str):
        """Sets the logical_rule of this Property/Relation.


        :param logical_rule: The logical_rule of this Property/Relation.  # noqa: E501
        :type: str
        """
        if not logical_rule:
            self._rest_model.advanced_config.logical_rule = None
            return
        if self._rest_model.advanced_config.logical_rule is None:
            self._rest_model.advanced_config.logical_rule = rest.LogicalRule()

        self._rest_model.advanced_config.logical_rule.content = logical_rule

    @property
    def alter_operation(self) -> AlterOperationEnum:
        """Gets the alter_operation of this Property/Relation.  # noqa: E501


        :return: The alter_operation of this Property/Relation.  # noqa: E501
        :rtype: AlterOperationEnum
        """
        alter_operation = self._rest_model.alter_operation
        return AlterOperationEnum(alter_operation) if alter_operation else None

    @alter_operation.setter
    def alter_operation(self, alter_operation: AlterOperationEnum):
        """Sets the alter_operation of this Property/Relation.


        :param alter_operation: The alter_operation of this Property/Relation.  # noqa: E501
        :type: AlterOperationEnum
        """
        self._rest_model.alter_operation = alter_operation

    def overwritten_by(self, other: Type["BaseProperty"]):
        """Overwrite all variables of the current class instance from another class instance."""
        import inspect

        members = inspect.getmembers(self.__class__)
        for name, member in members:
            if isinstance(member, property):
                if name == "sub_properties":
                    setattr(
                        self, name, [prop for _, prop in getattr(other, name).items()]
                    )
                else:
                    setattr(self, name, getattr(other, name))

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in self.__annotations__.items():
            if attr == "sub_properties":
                continue
            value = getattr(self, attr)
            if isinstance(value, typing.List):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, typing.Dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def to_rest(self):
        """Returns the REST model of this SpgType"""
        return self._rest_model

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, self.__class__):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, self.__class__):
            return True

        return self.to_dict() != other.to_dict()


class BaseSpgType(ABC):
    """Base class of `ConceptType`, `EntityType`, `EventType`, `StandardType`, `BasicType`."""

    _rest_model: Union[
        rest.ConceptType, rest.EntityType, rest.EventType, rest.StandardType
    ]

    def __init__(
        self,
        spg_type_enum=None,
        name=None,
        name_zh=None,
        desc=None,
        parent_type_name=None,
        properties=None,
        relations=None,
        **kwargs,
    ):
        if "rest_model" in kwargs:
            self._rest_model = kwargs["rest_model"]
        else:
            self._init_rest_model(
                spg_type_enum=spg_type_enum,
                name=name,
                name_zh=name_zh,
                desc=desc,
                parent_type_name=parent_type_name,
                properties=properties,
                relations=relations,
                **kwargs,
            )

    def _init_rest_model(self, **kwargs):
        """Init a BaseSpgType object."""
        super_klass = self.__class__.__name__
        self._rest_model = iter_init(getattr(rest, super_klass))
        for param, value in kwargs.items():
            setattr(self, param, value)

    @property
    def spg_type_enum(self) -> SpgTypeEnum:
        """Gets the spg_type_enum of this SpgType.  # noqa: E501


        :return: The spg_type_enum of this SpgType.  # noqa: E501
        :rtype: str
        """
        spg_type_enum = self._rest_model.spg_type_enum
        return SpgTypeEnum(spg_type_enum) if spg_type_enum else None

    @spg_type_enum.setter
    def spg_type_enum(self, spg_type_enum: SpgTypeEnum):
        """Sets the spg_type_enum of this SpgType.


        :param spg_type_enum: The spg_type_enum of this SpgType.  # noqa: E501
        :type: str
        """
        self._rest_model.spg_type_enum = spg_type_enum

    @property
    def name(self) -> str:
        """Gets the name of this SpgType.  # noqa: E501


        :return: The name of this SpgType.  # noqa: E501
        :rtype: str
        """
        return self._rest_model.basic_info.name.name

    @name.setter
    def name(self, name: str):
        """Sets the name of this SpgType.


        :param name: The name of this SpgType.  # noqa: E501
        :type: str
        """
        if name is None:  # noqa: E501
            raise ValueError(
                "Invalid value for `name`, must not be `None`"
            )  # noqa: E501

        if self._rest_model.basic_info.name.name != name:
            self._rest_model.basic_info.name.name = name

    @property
    def name_zh(self) -> str:
        """Gets the name_zh of this SpgType.  # noqa: E501


        :return: The name_zh of this SpgType.  # noqa: E501
        :rtype: str
        """
        return self._rest_model.basic_info.name_zh

    @name_zh.setter
    def name_zh(self, name_zh: str):
        """Sets the name_zh of this SpgType.


        :param name_zh: The name_zh of this SpgType.  # noqa: E501
        :type: str
        """

        if self._rest_model.basic_info.name_zh == name_zh:
            return
        self._rest_model.basic_info.name_zh = name_zh

    @property
    def desc(self) -> str:
        """Gets the desc of this SpgType.  # noqa: E501


        :return: The desc of this SpgType.  # noqa: E501
        :rtype: str
        """
        return self._rest_model.basic_info.desc

    @desc.setter
    def desc(self, desc: str):
        """Sets the desc of this SpgType.


        :param desc: The desc of this SpgType.  # noqa: E501
        :type: str
        """

        self._rest_model.basic_info.desc = desc

    @property
    def parent_type_name(self) -> str:
        """Gets the parent_type_name of this SpgType.  # noqa: E501


        :return: The parent_type_name of this SpgType.  # noqa: E501
        :rtype: str
        """
        return self._rest_model.parent_type_info.parent_type_identifier.name

    @parent_type_name.setter
    def parent_type_name(self, parent_type_name: str):
        """Sets the parent_type_name of this SpgType.


        :param parent_type_name: The parent_type_name of this SpgType.  # noqa: E501
        :type: BaseSpgType
        """
        if parent_type_name is None:
            return
        self._rest_model.parent_type_info.parent_type_identifier.name = parent_type_name

    @property
    def properties(self) -> Dict[str, Type["Property"]]:
        """Gets the properties of this SpgType.  # noqa: E501


        :return: The properties of this SpgType.  # noqa: E501
        :rtype: dict
        """
        from knext.schema.model.property import Property

        properties = {}
        for prop in self._rest_model.properties:
            properties[prop.basic_info.name.name] = Property(
                name=prop.basic_info.name.name,
                object_type_name=prop.object_type_ref.basic_info.name.name,
                rest_model=prop,
            )
        return properties

    @properties.setter
    def properties(self, properties: List[Type["Property"]]):
        """Sets the properties of this SpgType.


        :param properties: The properties of this SpgType.  # noqa: E501
        :type: list[Property]
        """
        if properties is None:
            return

        self._rest_model.properties = [prop.to_rest() for prop in properties]

    def add_property(self, prop: Type["Property"]):
        """Adds a property to this SpgType.


        :param prop: The property to add.  # noqa: E501
        :type: Property
        """
        prop.alter_operation = AlterOperationEnum.Create
        self._rest_model.properties.append(prop.to_rest())
        return self

    @property
    def relations(self) -> Dict[str, Type["Relation"]]:
        """Gets the relations of this SpgType.  # noqa: E501


        :return: The relations of this SpgType.  # noqa: E501
        :rtype: dict
        """
        from knext.schema.model.relation import Relation

        relations = {}
        for relation in self._rest_model.relations:
            predicate_name = relation.basic_info.name.name
            object_type_name = relation.object_type_ref.basic_info.name.name
            relations[predicate_name + "_" + object_type_name] = Relation(
                name=predicate_name,
                object_type_name=object_type_name,
                rest_model=relation,
            )
        return relations

    @relations.setter
    def relations(self, relations: List["Relation"]):
        """Sets the relations of this SpgType.


        :param relations: The relations of this SpgType.  # noqa: E501
        :type: list[Relation]
        """

        if relations is None:
            return

        self._rest_model.relations = [relation.to_rest() for relation in relations]

    def add_relation(self, relation: Type["Relation"]):
        """Adds a relation to this SpgType.


        :param relation: The relation to add.  # noqa: E501
        :type: Relation
        """

        relation.alter_operation = AlterOperationEnum.Create
        self._rest_model.relations.append(relation.to_rest())
        return self

    @property
    def alter_operation(self) -> Optional[AlterOperationEnum]:
        """Gets the alter_operation of this SpgType.  # noqa: E501


        :return: The alter_operation of this SpgType.  # noqa: E501
        :rtype: AlterOperationEnum
        """
        alter_operation = self._rest_model.alter_operation
        return AlterOperationEnum(alter_operation) if alter_operation else None

    @alter_operation.setter
    def alter_operation(self, alter_operation: AlterOperationEnum):
        """Sets the alter_operation of this SpgType.


        :param alter_operation: The alter_operation of this SpgType.  # noqa: E501
        :type: AlterOperationEnum
        """
        self._rest_model.alter_operation = alter_operation

    @staticmethod
    def by_type_enum(type_enum: str):
        """Reflection from type enum to subclass object of BaseSpgType."""

        import knext.schema.model.spg_type as spg_type

        class_obj = getattr(spg_type, f"{SpgTypeEnum(type_enum).name}Type")
        return class_obj

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in self.__annotations__.items():
            value = getattr(self, attr)
            if isinstance(value, typing.List):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, typing.Dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def to_rest(self):
        """Returns the REST model of this SpgType"""
        return self._rest_model

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, self.__class__):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, self.__class__):
            return True

        return self.to_dict() != other.to_dict()



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from typing import List, Dict

from knext.schema import rest
from knext.common.base.client import Client
from knext.schema.model.base import BaseSpgType, AlterOperationEnum, SpgTypeEnum
from knext.schema.model.relation import Relation


class SchemaClient(Client):
    """ """

    _rest_client = rest.SchemaApi()

    def __init__(self, host_addr: str = None, project_id: int = None):
        super().__init__(host_addr, project_id)

        self._session = None

    def query_spg_type(self, spg_type_name: str) -> BaseSpgType:
        """Query SPG type by name."""
        rest_model = self._rest_client.schema_query_spg_type_get(spg_type_name)
        type_class = BaseSpgType.by_type_enum(f"{rest_model.spg_type_enum}")

        if rest_model.spg_type_enum == SpgTypeEnum.Concept:
            return type_class(
                name=spg_type_name,
                hypernym_predicate=rest_model.concept_layer_config.hypernym_predicate,
                rest_model=rest_model,
            )
        else:
            return type_class(name=spg_type_name, rest_model=rest_model)

    def query_relation(
        self, subject_name: str, predicate_name: str, object_name: str
    ) -> Relation:
        """Query relation type by s_p_o name."""
        rest_model = self._rest_client.schema_query_relation_get(
            subject_name, predicate_name, object_name
        )
        return Relation(
            name=predicate_name, object_type_name=object_name, rest_model=rest_model
        )

    def create_session(self):
        """Create session for altering schema."""
        return self.SchemaSession(self._rest_client, self._project_id)

    class SchemaSession:
        def __init__(self, client, project_id):
            self._alter_spg_types: List[BaseSpgType] = []
            self._rest_client = client
            self._project_id = project_id

            self._spg_types = {}
            self.__spg_types = {}
            self._init_spg_types()

        def _init_spg_types(self):
            """Query project schema and init SPG types in session."""
            project_schema = self._rest_client.schema_query_project_schema_get(
                self._project_id
            )
            for spg_type in project_schema.spg_types:
                spg_type_name = spg_type.basic_info.name.name
                type_class = BaseSpgType.by_type_enum(spg_type.spg_type_enum)
                if spg_type.spg_type_enum == SpgTypeEnum.Concept:
                    self._spg_types[spg_type_name] = type_class(
                        name=spg_type_name,
                        hypernym_predicate=spg_type.concept_layer_config.hypernym_predicate,
                        rest_model=spg_type,
                    )
                else:
                    self._spg_types[spg_type_name] = type_class(
                        name=spg_type_name, rest_model=spg_type
                    )

        @property
        def spg_types(self) -> Dict[str, BaseSpgType]:
            return self._spg_types

        def get(self, spg_type_name) -> BaseSpgType:
            """Get SPG type by name from project schema."""
            spg_type = self._spg_types.get(spg_type_name)
            if spg_type is None:
                spg_type = self.__spg_types.get(spg_type_name)
                if spg_type is None:
                    raise ValueError(f"{spg_type_name} is not existed")
                else:
                    return self.__spg_types.get(spg_type_name)
            return self._spg_types.get(spg_type_name)

        def create_type(self, spg_type: BaseSpgType):
            """Add an SPG type in session with `CREATE` operation."""
            spg_type.alter_operation = AlterOperationEnum.Create
            self.__spg_types[spg_type.name] = spg_type
            self._alter_spg_types.append(spg_type)
            return self

        def update_type(self, spg_type: BaseSpgType):
            """Add an SPG type in session with `UPDATE` operation."""
            spg_type.alter_operation = AlterOperationEnum.Update
            self._alter_spg_types.append(spg_type)
            return self

        def delete_type(self, spg_type: BaseSpgType):
            """Add an SPG type in session with `DELETE` operation."""
            spg_type.alter_operation = AlterOperationEnum.Delete
            self._alter_spg_types.append(spg_type)
            return self

        def commit(self):
            """Commit all altered schemas to server."""
            schema_draft = []
            for spg_type in self._alter_spg_types:
                for prop in spg_type.properties.values():
                    if prop.object_spg_type is None:
                        object_spg_type = self.get(prop.object_type_name)
                        prop.object_spg_type = object_spg_type.spg_type_enum
                    for sub_prop in prop.sub_properties.values():
                        if sub_prop.object_spg_type is None:
                            object_spg_type = self.get(sub_prop.object_type_name)
                            sub_prop.object_spg_type = object_spg_type.spg_type_enum
                for rel in spg_type.relations.values():
                    if rel.is_dynamic is None:
                        rel.is_dynamic = False
                    if rel.object_spg_type is None:
                        object_spg_type = self.get(rel.object_type_name)
                        rel.object_spg_type = object_spg_type.spg_type_enum
                    for sub_prop in rel.sub_properties.values():
                        if sub_prop.object_spg_type is None:
                            object_spg_type = self.get(sub_prop.object_type_name)
                            sub_prop.object_spg_type = object_spg_type.spg_type_enum
                schema_draft.append(spg_type.to_rest())
            if len(schema_draft) == 0:
                return

            request = rest.SchemaAlterRequest(
                project_id=self._project_id, schema_draft=rest.SchemaDraft(schema_draft)
            )
            self._rest_client.schema_alter_schema_post(schema_alter_request=request)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


from abc import ABC
from enum import Enum
from typing import List, Union, Type

import networkx as nx

from knext.common.base.restable import RESTable
from knext.common.base.runnable import Runnable


class ComponentTypeEnum(str, Enum):
    Builder = "BUILDER"
    Reasoner = "REASONER"
    GraphLearning = "GRAPH_LEARNING"


class Component(Runnable, RESTable, ABC):
    """
    Base class for all component.
    """

    @property
    def id(self):
        return str(id(self))

    @property
    def type(self):
        return

    @property
    def label(self):
        return

    @property
    def name(self):
        return self.__class__.__name__

    def to_dict(self):
        return {"id": self.id, "name": self.name}

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __rshift__(
        self,
        other: Union[
            Type["Chain"],
            List[Type["Chain"]],
            Type["Component"],
            List[Type["Component"]],
            None,
        ],
    ):
        from knext.common.base.chain import Chain
        from knext.builder.component.mapping import SPGTypeMapping
        from knext.builder.component.mapping import _SPGTypeMappings

        if not other:
            return self
        if not isinstance(other, list):
            other = [other]
        dag_list = []
        if all(isinstance(o, SPGTypeMapping) for o in other):
            other = [_SPGTypeMappings(spg_type_mappings=other)]
        for o in other:
            if not o:
                dag = nx.DiGraph()
                self.last = True
                dag.add_node(self)
                dag_list.append(dag)
            if isinstance(o, Component):
                dag = nx.DiGraph()
                dag.add_node(self)
                dag.add_node(o)
                dag.add_edge(self, o)
                dag_list.append(dag)
            elif isinstance(o, Chain):
                dag = nx.DiGraph()
                dag.add_node(self)
                end_nodes = [
                    node
                    for node, out_degree in dag.out_degree()
                    if out_degree == 0 or node.last
                ]
                start_nodes = [
                    node for node, in_degree in o.dag.in_degree() if in_degree == 0
                ]

                if len(end_nodes) > 0 and len(start_nodes) > 0:
                    for end_node in end_nodes:
                        for start_node in start_nodes:
                            combined_dag.add_edge(end_node, start_node)
                combined_dag = nx.compose(dag, o.dag)
                dag_list.append(combined_dag)
        final_dag = nx.compose_all(dag_list)

        return Chain(dag=final_dag)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import click

from knext.command.sub_command.builder import execute_job
from knext.command.sub_command.config import edit_config
from knext.command.sub_command.config import list_config
from knext.command.sub_command.project import create_project
from knext.command.sub_command.project import list_project
from knext.command.sub_command.reasoner import execute_reasoner_job
from knext.command.sub_command.schema import commit_schema
from knext.command.sub_command.schema import list_schema
from knext.command.sub_command.schema import reg_concept_rule
from knext.command.exception import _ApiExceptionHandler
from knext import __version__


@click.group(cls=_ApiExceptionHandler)
@click.version_option(__version__)
def _main() -> None:
    pass


@_main.group()
def config() -> None:
    """Knext config."""
    pass


config.command("list")(list_config)
config.command("set")(edit_config)


@_main.group()
def builder() -> None:
    """Builder client."""
    pass


builder.command("execute")(execute_job)


@_main.group()
def project() -> None:
    """Project client."""
    pass


project.command("create")(create_project)
project.command("list")(list_project)


@_main.group()
def schema() -> None:
    """Schema client."""
    pass


schema.command("commit")(commit_schema)
schema.command("list")(list_schema)
schema.command("reg_concept_rule")(reg_concept_rule)


@_main.group()
def reasoner() -> None:
    """Reasoner client."""
    pass


reasoner.command("execute")(execute_reasoner_job)

if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import pprint
from typing import Dict, Any, List, Tuple
from knext.schema.model.schema_helper import (
    SPGTypeName,
    PropertyName,
    RelationName,
)


class SPGRecord:
    """Data structure in operator, used to store entity information."""

    def __init__(self, spg_type_name: SPGTypeName):
        self._spg_type_name = spg_type_name
        self._properties = {}
        self._relations = {}

    @property
    def spg_type_name(self) -> SPGTypeName:
        """Gets the spg_type_name of this SPGRecord.  # noqa: E501


        :return: The spg_type_name of this SPGRecord.  # noqa: E501
        :rtype: str
        """
        return self._spg_type_name

    @spg_type_name.setter
    def spg_type_name(self, spg_type_name: SPGTypeName):
        """Sets the spg_type_name of this SPGRecord.


        :param spg_type_name: The spg_type_name of this SPGRecord.  # noqa: E501
        :type: str
        """
        self._spg_type_name = spg_type_name

    @property
    def properties(self) -> Dict[PropertyName, str]:
        """Gets the properties of this SPGRecord.  # noqa: E501


        :return: The properties of this SPGRecord.  # noqa: E501
        :rtype: dict
        """
        return self._properties

    @properties.setter
    def properties(self, properties: Dict[PropertyName, str]):
        """Sets the properties of this SPGRecord.


        :param properties: The properties of this SPGRecord.  # noqa: E501
        :type: dict
        """
        self._properties = properties

    @property
    def relations(self) -> Dict[str, str]:
        """Gets the relations of this SPGRecord.  # noqa: E501


        :return: The relations of this SPGRecord.  # noqa: E501
        :rtype: dict
        """
        return self._relations

    @relations.setter
    def relations(self, relations: Dict[str, str]):
        """Sets the properties of this SPGRecord.


        :param relations: The relations of this SPGRecord.  # noqa: E501
        :type: dict
        """
        self._relations = relations

    def get_property(
        self, property_name: PropertyName, default_value: str = None
    ) -> str:
        """Gets a property of this SPGRecord by name.  # noqa: E501


        :param property_name: The property name.  # noqa: E501
        :param default_value: If property value is None, the default_value will be return.  # noqa: E501
        :return: A property value.  # noqa: E501
        :rtype: str
        """
        return self.properties.get(property_name, default_value)

    def upsert_property(self, property_name: PropertyName, value: str):
        """Upsert a property of this SPGRecord.  # noqa: E501


        :param property_name: The updated property name.  # noqa: E501
        :param value: The updated property value.  # noqa: E501
        :type: str
        """
        self.properties[property_name] = value
        return self

    def upsert_properties(self, properties: Dict[PropertyName, str]):
        """Upsert properties of this SPGRecord.  # noqa: E501


        :param properties: The updated properties.  # noqa: E501
        :type: dict
        """
        self.properties.update(properties)
        return self

    def remove_property(self, property_name: PropertyName):
        """Removes a property of this SPGRecord.  # noqa: E501


        :param property_name: The property name.  # noqa: E501
        :type: str
        """
        self.properties.pop(property_name)
        return self

    def remove_properties(self, property_names: List[PropertyName]):
        """Removes properties by given names.  # noqa: E501


        :param property_names: A list of property names.  # noqa: E501
        :type: list
        """
        for property_name in property_names:
            self.properties.pop(property_name)
        return self

    def get_relation(
        self,
        relation_name: RelationName,
        object_type_name: SPGTypeName,
        default_value: str = None,
    ) -> str:
        """Gets a relation of this SPGRecord by name.  # noqa: E501


        :param relation_name: The relation name.  # noqa: E501
        :param object_type_name: The object SPG type name.  # noqa: E501
        :param default_value: If property value is None, the default_value will be return.  # noqa: E501
        :return: A relation value.  # noqa: E501
        :rtype: str
        """
        return self.relations.get(relation_name + "#" + object_type_name, default_value)

    def upsert_relation(
        self, relation_name: RelationName, object_type_name: SPGTypeName, value: str
    ):
        """Upsert a relation of this SPGRecord.  # noqa: E501


        :param relation_name: The updated relation name.  # noqa: E501
        :param object_type_name: The object SPG type name.  # noqa: E501
        :param value: The updated relation value.  # noqa: E501
        :type: str
        """
        self.relations[relation_name + "#" + object_type_name] = value
        return self

    def upsert_relations(self, relations: Dict[Tuple[RelationName, SPGTypeName], str]):
        """Upsert relations of this SPGRecord.  # noqa: E501


        :param relations: The updated relations.  # noqa: E501
        :type: dict
        """
        for (relation_name, object_type_name), value in relations.items():
            self.relations[relation_name + "#" + object_type_name] = value
        return self

    def remove_relation(
        self, relation_name: RelationName, object_type_name: SPGTypeName
    ):
        """Removes a relation of this SPGRecord.  # noqa: E501


        :param relation_name: The relation name.  # noqa: E501
        :param object_type_name: The object SPG type name.  # noqa: E501
        :type: str
        """
        self.relations.pop(relation_name + "#" + object_type_name)
        return self

    def remove_relations(self, relation_names: List[Tuple[RelationName, SPGTypeName]]):
        """Removes relations by given names.  # noqa: E501


        :param relation_names: A list of relation names.  # noqa: E501
        :type: list
        """
        for (relation_name, object_type_name) in relation_names:
            self.relations.pop(relation_name + "#" + object_type_name)
        return self

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.__dict__())

    def to_dict(self):
        """Returns the model properties as a dict"""

        return {
            "spgTypeName": self.spg_type_name,
            "properties": {
                **self.properties,
                **self.relations,
            },
        }

    def __dict__(self):
        """Returns this SPGRecord as a dict"""
        return {
            "spgTypeName": self.spg_type_name,
            "properties": self.properties,
            "relations": self.relations,
        }

    @classmethod
    def from_dict(cls, input: Dict[str, Any]):
        """Returns the model from a dict"""
        spg_type_name = input.get("spgTypeName")
        _cls = cls(spg_type_name)
        properties = input.get("properties")
        for k, v in properties.items():
            if "#" in k:
                relation_name, object_type_name = k.split("#")
                _cls.relations.update({relation_name + "#" + object_type_name: v})
            else:
                _cls.properties.update({k: v})

        return _cls

    def __repr__(self):
        """For `print` and `pprint`"""
        return pprint.pformat(self.__dict__())



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import os
from abc import ABC
from typing import Dict, Any, Type

from knext.builder import rest


class BaseOp(ABC):
    """Base class for all user-defined operator functions.

    The execution logic of the operator needs to be implemented in the `eval` method.
    """

    """Operator name."""
    name: str
    """Operator description."""
    desc: str = ""
    """Operator params."""
    params: Dict[str, str] = None

    _registry = {}
    _local_path: str
    _module_path: str
    _version: int
    _has_registered: bool = False

    def __init__(self, params: Dict[str, str] = None):
        self.params = params

    def invoke(self, *args):
        """Used to implement operator execution logic."""
        raise NotImplementedError(
            f"{self.__class__.__name__} need to implement `invoke` method."
        )

    def _handle(self, *inputs) -> Dict[str, Any]:
        """Only available for Builder in OpenSPG to call through the pemja tool."""
        pre_input = self._pre_process(*inputs)
        output = self.invoke(*pre_input)
        post_output = self._post_process(output)
        return post_output

    @staticmethod
    def _pre_process(*inputs):
        """Convert data structures in building job into structures in operator before `eval` method."""
        return inputs

    @staticmethod
    def _post_process(output) -> Dict[str, Any]:
        """Convert result structures in operator into structures in building job after `eval` method."""
        return output.to_dict()

    @classmethod
    def register(cls, name: str, local_path: str, module_path: str):
        """
        Register a class as subclass of BaseOp with name and local_path.
        After registration, the subclass object can be inspected by `BaseOp.by_name(op_name)`.
        """

        def add_subclass_to_registry(subclass: Type["BaseOp"]):
            subclass.name = name
            subclass._local_path = local_path
            subclass._module_path = module_path
            if name in cls._registry:
                raise ValueError(
                    f"Operator [{name}] conflict in {subclass._local_path} and {cls.by_name(name)._local_path}."
                )
            cls._registry[name] = subclass
            if hasattr(subclass, "bind_to"):
                subclass.__bases__[0].bind_schemas[subclass.bind_to] = name
            return subclass

        return add_subclass_to_registry

    @classmethod
    def by_name(cls, name: str):
        """Reflection from op name to subclass object of BaseOp."""
        if name in cls._registry:
            subclass = cls._registry[name]
            return subclass
        else:
            raise ValueError(f"{name} is not a registered name for {cls.__name__}. ")

    def to_rest(self):
        if not hasattr(self, "_local_path"):
            import inspect

            self._local_path = inspect.getfile(self.__class__)
        if not hasattr(self, "name"):
            self.name = self.__class__.__name__
        if not hasattr(self, "_module_path"):
            self._module_path = os.path.splitext(os.path.basename(self._local_path))[0]
        return rest.OperatorConfig(
            file_path=self._local_path,
            module_path=self._module_path,
            class_name=self.name,
            method="_handle",
            params=self.params,
        )

    @property
    def has_registered(self):
        return self._has_registered
